#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
(note, this is generated from the command line version, 80% representative
of the options for the direct function call)
optional arguments:
  -h, --help            show this help message and exit

  data.file             Specify data file. This file may be .h5, .hdf, .mat, .csv, .txt.
                        Can specify or override the filetype and internal format using a
                        namespace syntax '::'.
                         - "xfer.___::hdf" for hdf5 with an unrecognized suffix
                         - "xfer.out::csv::,F,Xr,Xi,W" to interpret as csv with 4 columns including
                           data real and imaginary parts
                         - "xfer.out::csv::-F,Xm,Xp,W" to interpret as csv with 4 columns whitespace
                           separated including data magnitude and phase [radians]
                        all of the CSV column specifiers are:
                         - first specifier is the delimiter, one of any puncuation such as  ",:|".
                           the symbol '-' indicates to use whitespace of any kind.
                         - F, f for frequency in [Hz]
                         - Xr, Xi for xfer real and imaginary parts
                         - Xm, Xa for xfer magnitude or equivalently amplitude
                         - Xdb, XdB xfer in dBs
                         - Xp for xfer phase in radians
                         - Xd for xfer phase in degrees
                         - S, W, w for weights or SNR
                         - E, e for emphasis
                         - . for ignore column
                        by default, the CSV is expected to be formatted as ",XrXiFSE". If columns
                        are missing they are dropped
                        matlab and h5 structures are assumed to have keys:
                         'F_Hz', 'xfer', 'SNR', 'emphasis',
                        but arguments in the "data" group can override these keys, or specify
                        columns.
  output.file           Specify the output file to store fit results including zpk of the
                        chosen and alternative fits, the configurations, and versioning information.
                        Possible output extensions are .h5, .hdf, .mat, .pkl, .json, .yaml. Binary
                        formats .mat, .h5, .pkl will include the original data, for full reconstruction
                        of the fit. wield.iirrational may be called on the output file to rerun the fit.
  -c CONFIG, --config CONFIG
                        Specify configuration file. This file may be .json, .yaml, .ini, .h5, .mat.
                        These will be interpreted as a dictionary or structure full of keyword
                        arguments. This may be specified multiple times to overlay settings.
                        Additional command line setting always override these configuration files.
  -L {-,none,Sf}, --LIGO_foton {-,none,Sf}
                        What type of foton output would you prefer?
                          - '-' or 'None' to suppress foton output
                          - Sf for frequency domain (default)
                          - n for normalized (not yet supported)
                          - Sw for radial (not yet supported)
  -j CHOOSE, --choose CHOOSE
                        How to decide the optimal order. Special parameters are:
                         - "prompt", "shell", "interactive" (the default) To enter a command prompt that plots and requests
                           a choice
                         - "auto10" the default choice to use the shell if the baseline order is
                           above 10 unless there is no choice. (10 may be changed)
                         - "baseline" to use the baseline fit before the ChiSq starts rising.
                         - "baseline10" to use the baseline "best fit" unless it is over order 10
                                        (10 can be configured to other integers)
                         - integer. To force a choice of order. It will always use the best fit
                           under this order.
  -m {full,fit,reduce,rational,dumpargs,dumpargs_full,printdata,printconfig,printsettings}, --mode {full,fit,reduce,rational,dumpargs,dumpargs_full,printdata,printconfig,printsettings}
                        Fitting mode to use, to change the automation level. Must be one of
                         - "full": to use rational fitting to get initial parameters, then alternate
                                   optimization and order reduction
                         - "rational": to use rational fitting to get initial parameters and then
                                       stop after the simplest order reduction. Useful to use with
                                       other fitters. Does not perform delay fitting.
                         - "fit": To take an initial ZPK guess, and only fit/optimize it. Good
                                  for refining previous fits.
                         - "reduce": To take an initial ZPK guess, and then alternate fitting and
                                     order reduction.
                         - "dumpargs": Dump arguments to determine current settings
                         - "dumpargs_full": Dump arguments and all generated settings to see the full run setup
                         - "printdata": Dump the layout of the data file specified
                         - "printconfig": dump the layout of the config files specified
                         - "printsettings": dump the fully loaded settings specified
  -N F_NYQUIST_HZ, --F_nyquist_Hz F_NYQUIST_HZ
                        This selects the Nyquist frequency for Z-domain fitting. If None or not
                        specified, the fits are done in the "Sf" domain - the s domain scaled for
                        units of frequency.
  -I INFORMATION, --information INFORMATION
                        A string to detail what this fit is of. For documentation purposes.
  --overwrite           Allow the output file to overwrite the previous output file (be careful).

groups:
  The data and config file formats (except csv) all load internally to a
  dictionary or structure representation. These arguments specify the
  keys or groups storing the relevant data or configuration subkeys. These
  also control the output groups when saving files.

  -C CONFIG_GROUP, --config_group CONFIG_GROUP
                        Group(s) within the loaded config or data file to search for
                        configuration elements. May be a comma separated list to aggregate groups.
                        defaults to "config,conf,IIRconfig,IIRconf". The first element will be
                        the group where configurations are stored to the output file.
  -D DATA_GROUP, --data_group DATA_GROUP
                        Group(s) within the loaded data file to search for data elements
                        (see "data" section). May be a comma separated list to aggregate groups.
                        defaults to "data,IIRdata,xfer". The first specified element will be
                        the group that data is stored to the output file.

data:
  Specify keys to search within the data group for specific arrays.
  -F,   --frequency,
  -Xc,  --dataXcomplex
  -Xr,  --dataXreal,      -Xi,  --dataXimaginary
  -Xa,  --dataXamplitude, -Xm,  --dataXmagnitude
  -XdB, --dataXdB,        -Xdb, --dataXdb
  -Xd,  --dataXphaseDeg,  -Xp,  --dataXphaseRad
  -S,   --dataSNR,        -W,   --dataW
  -E,   --dataEmphasis

order:
  --relative_degree RELATIVE_DEGREE
                        Sets the initial relative degree (number zeros minus polse).
                        Defaults to None, which will be the midpoint of the min and max if they
                        are specified, otherwise this will default to 0, which typically will
                        still fit filters to the correct degree. If constrained to be different
                        than the data, the filter will enter the asymptotic regime set by the
                        relative degree within 2x of the 'root_bandwidth_Hz_max' setting.
  --relative_degree_max RELATIVE_DEGREE_MAX
                        Maximum value for the filter relative degree (number zeros minus polse).
                        Defaults to the relative degree (which may be None). If this value is
                        None, the degree is unconstrained and the fit will land at some degree that fits well.
  --relative_degree_min RELATIVE_DEGREE_MIN
                        Minimum value for the filter relative degree (number zeros minus polse).
                        Defaults to the relative degree (which may be None). If this value is
                        None, the degree is unconstrained and the fit will land at some degree that fits well.
  --total_degree_min TOTAL_DEGREE_MIN
                        Minimum degree to search through during the successive order reduction phase.
                        Defaults to 2. If None, then successive reduction will not be performed.
  --order_initial ORDER_INITIAL
                        Order to use in rational fitting. If not specified, the order is
                        increased until the residuals plateau.
  --order_max ORDER_MAX
                        Maximium order that the rational fitter order estimation may use.
                        Increasing this value from the default may impact numerical stability.
  --order_min ORDER_MIN
                        Minimum order to use during order estimation. Smaller values will speed
                        up fits, but may fail to fix complex data.

plots:
  -r PLOT_FIT, --plot_fit PLOT_FIT
                        filename to plot (review) the chosen fit to. May be any extension supported by matplotlib.
                        [.pdf, .svg, .png, .jpg, ...]
  -R PLOT_ORDER, --plot_order PLOT_ORDER
                        filename to plot (Review) potential orders and residuals to. May be any extension supported by matplotlib.
                        [.pdf, .svg, .png, .jpg, ...]

delay:
  --delay_s_max DELAY_S_MAX
                        The maximum delay in seconds. Defaults to None, in which case delay is
                        never fit as a free parameter.
  --delay_s DELAY_S     Use this delay (in seconds) for the initial fitting up until the
                        "baseline" fit determination is complete. By default this is the same
                        as delay_s_min, unless delay_s_min is negative, in which case it defaults
                        to the smaller of 0 seconds or delay_s_max.
  --delay_s_min DELAY_S_MIN
                        The minimum delay in seconds. Defaults to 0. Also sets the typical
                        default value of the delay used for the initial part of the fits.
                        Must always be specified (cannot be None).

logging:
  -l LOG_LEVEL, --log_level LOG_LEVEL
                        Log level default for all logging types
  --log_level_alert LOG_LEVEL_ALERT
                        Data on useful statistical tests, particularly those which cause input
                        data/SNR/emphasis to be reinterpreted. Typically logs at level 2-4.
  --log_level_debug LOG_LEVEL_DEBUG
                        Debugging reports used for development.
  --log_level_info LOG_LEVEL_INFO
                        Logging on miscellaneous details.
  --log_level_rationale LOG_LEVEL_RATIONALE
                        Detailed explanations of tests and algorithms. Extremely verbose and
                        intended for first users and digest reports.
  --log_level_warn LOG_LEVEL_WARN
                        Warnings about failed tests, or data operating in a regime unexpected to
                        work or that validation should be made.

SNR Adjustments:
  --SNR_cuttoff SNR_CUTTOFF
                        Maximum SNR. Hard cutoff for SNR above this level.
  --SNR_estimate_width SNR_ESTIMATE_WIDTH
                        The window width of nearby points to use for averaging and median filtering
                        when estimating the SNR via windowed sample variance. Default is 10.
                        None or 0 indicates to not try.
  --SNR_regularize_ratio SNR_REGULARIZE_RATIO
                        Ratio of effective data points, determined by the dynamic range of the
                        SNR weighting, to the actual number of data points. This parameter
                        adjusts how the SNR is adjusted to ensure sufficient data points for a
                        good fit.
  --SNR_regularize_scale SNR_REGULARIZE_SCALE
                        Similar to SNR_regularize_ratio, but instead of a direct ratio, the ratio
                        is determined as 'SNR_regularize_scale' / max(SNR). For very high SNR
                        fits, the SNR should be whitened significantly using the ratio adjustment,
                        since the error is dominated by systematics rather than statistical noise.
                        The default is 10, causing a %90 coverage ratio when the highest SNR is
                        100 (1% statistical error).
  --trust_SNR           Overall parameter determining if SNR/W statistical weighting input should
                        be trusted. If False (the default) then additional statistical tests are
                        run to adjust the SNR for better fits. If the user can fully trust their
                        SNR estimators and fits to be unbiased, then the user likely does not
                        need this tool.

advanced:
  --resavg_RthreshOrdC RESAVG_RTHRESHORDC
                        The threshold relative change in the average residuals to accept a fit
                        of equal order. Only used for the "baseline" fit determination before
                        delay is activated and not used during the total order reduction.
  --resavg_RthreshOrdDn RESAVG_RTHRESHORDDN
                        The threshold relative change in the average residuals to accept a fit
                        of lower order. Only used for the "baseline" fit determination before
                        delay is activated and not used during the total order reduction.
  --resavg_RthreshOrdUp RESAVG_RTHRESHORDUP
                        The threshold relative change in the average residuals to accept a fit
                        of higher order. Only used for the "baseline" fit determination before
                        delay is activated and not used during the total order reduction.

emphasis:
  --emphasis_order_initial EMPHASIS_ORDER_INITIAL
                        Order to use in rational fitting during the emphasis application stage.
  --emphasis_order_max EMPHASIS_ORDER_MAX
                        Maximium order that the rational fitter order estimation may use during
                        the emphasis application stage.
  --emphasis_order_min EMPHASIS_ORDER_MIN
                        Minimum order to use during order estimation during the emphasis application
                        stage.

fitting:
  -k GAIN, --gain GAIN  Initial gain for the fit
  -p [POLES [POLES ...]], --poles [POLES [POLES ...]]
                        Initial poles used in the fit.
  -P [POLES_OVERLAY [POLES_OVERLAY ...]], --poles_overlay [POLES_OVERLAY [POLES_OVERLAY ...]]
                        poles guaranteed to be part of the fit.
  -z [ZEROS [ZEROS ...]], --zeros [ZEROS [ZEROS ...]]
                        Initial zeros used in the fit.
  -Z [ZEROS_OVERLAY [ZEROS_OVERLAY ...]], --zeros_overlay [ZEROS_OVERLAY [ZEROS_OVERLAY ...]]
                        zeros guaranteed to be part of the fit.

operating mode:
  --suggest             How to use the rational fitter when provided an initial ZPK argument.
                        the default of False causes the rational to overlay the ZPK suggestion,
                        forcing the roots in the suggestion to be used during the optimization
                        stage. If set to True, the ZPK will suggest initial poles to use during
                        the fit, which can accellerate convergence or cause it to start at a
                        lower initial order, speeding up future fitting operations.

residuals:
  --alternate_residuals ALTERNATE_RESIDUALS
                        Flag to switch from using log/phase residuals to using dual
                        ratio residuals as the principle fit residuals. Log/phase is typically
                        better behaved.
  --residuals_type {log,dualA,dualB,poles,zeros}
                        Standard residuals type to use for the optimizations. Must be one of the
                        definitions below, where R=fit/data, the ratio of fit to data and W is
                        the SNR:
                            log:   W*(ln(|R|) + 1j*R.imag/|R|)
                            dualA: W*(R + 1/R - 2)/2
                            dualB: W*(R - 1/R)/2
                            poles: W*(1/R - 1)
                            zeros: W*(R - 1)
  --residuals_type_alt {log,dualA,dualB,poles,zeros}
                        Standard alternate residuals type to use during optimization annealing.
                        See residuals_type for options.

speed:
  --greedy_order GREEDY_ORDER
                        Does only partial optimization during the reduce step. Greatly speeds
                        up order reduction, but may not be as effective. Defaults to 30, where
                        it uses greedy optimization until reaching order 30, then uses combinatoric
                        optimization.

tunings:
  --distance_limit_scale DISTANCE_LIMIT_SCALE
                        Scaling for how to limit root bandwidth for roots in-between data points.
                        Increase to force roots to lower Q, decrease to allow higher Qs.
  --root_bandwidth_Hz_max ROOT_BANDWIDTH_HZ_MAX
                        Maximum bandwidth of any pole or zero in the fit. This affects the
                        asymptotic rolloff when relative_degree is constrained.
"""
# this file is auto generated during the setup.py build phase.



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

import os.path as path

from wield.iirrational.testing import IIRrational_data
from wield.iirrational import v1
from wield.iirrational.annotate.exporters import npm_markdown_pdf


def test_simple_digest(browser, plotsections, plot_verbosity):
    data_name = "simple2"
    out = v1.data2filter(
        IIRrational_data(
            data_name,
            instance_num=1,
            set_num=2,
        ),
        order_initial=30,
        delay_s=0,
    )

    folder = path.join(path.split(__file__)[0], "out-" + data_name)
    print("FOLDER", folder)

    if browser:
        out.digest_write(
            folder,
            md_name="browser.md",
            plot_verbosity=10,
            clear_plots=True,
            regex_plotsections=plotsections,
            plot_format="png",
            dpi=100,
        )
    if plot_verbosity is None:
        plot_verbosity = 4
    out.digest_write(
        folder,
        plot_verbosity=plot_verbosity,
        clear_plots=True,
        regex_plotsections=plotsections,
        plot_format="png",
        dpi=200,
        exporter=npm_markdown_pdf("digest.pdf"),
        MP_workers=1,
    )
    return


if __name__ == "__main__":
    test_simple1(False, None, 10)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
Some utilities to help with dataset creation and annotation
"""
from wield.bunch import Bunch

import numpy as np

from .. import annotate
from .. import TFmath
from .. import representations
from .. import fitters_ZPK

from . import plots


def make_description(
    generator, sets=1, instances=1, description=None, annotation=None, **kwargs
):
    doc = generator.__doc__
    if description is None:
        if doc is not None:
            description = annotate.padding_remove(doc)
        else:
            description = "<no description>"
    else:
        description = annotate.padding_remove(description)
        if annotation is None and doc is not None:
            annotation = annotate.padding_remove(doc)
    return Bunch(
        generator=generator,
        instances=instances,
        description=description,
        annotation=annotation,
        **kwargs
    )


def generator_autofill(
    F_Hz,
    SNR,
    F_nyquist_Hz,
    data=None,
    bestfit_ZPK_z=None,
    bestfit_ZPK_s=None,
    delay_s=0,
    residuals_log_best=None,
    sN=0,
    iN=0,
    **kwargs
):
    if bestfit_ZPK_s:
        # replace with the ZPKs filled out fully
        bestfit_ZPK_s = TFmath.ZPK_fill(bestfit_ZPK_s)

    if bestfit_ZPK_z:
        # replace with the ZPKs filled out fully
        bestfit_ZPK_z = TFmath.ZPK_fill(bestfit_ZPK_z)

    if not bestfit_ZPK_z and F_nyquist_Hz is not None:
        if bestfit_ZPK_s:
            bestfit_ZPK_z = TFmath.StoZ(
                bestfit_ZPK_s,
                F_nyquist_Hz=F_nyquist_Hz,
            )

    if not bestfit_ZPK_s and F_nyquist_Hz is not None:
        if bestfit_ZPK_z:
            bestfit_ZPK_s = TFmath.ZtoS(
                bestfit_ZPK_z,
                F_nyquist_Hz=F_nyquist_Hz,
            )

    if data is None:
        if SNR is not None:
            rand = np.random.RandomState()
            rand.seed(iN)
            N = len(F_Hz)
            rel_noise = rand.normal(1, 1 / SNR, N) + 1j * rand.normal(0, 1 / SNR, N)
        else:
            rel_noise = 1

        if bestfit_ZPK_s:
            data = rel_noise * TFmath.TF_ZPK(
                F_Hz,
                ZPK=bestfit_ZPK_s,
            )
        elif bestfit_ZPK_z:
            data = rel_noise * TFmath.TF_ZPK(
                F_Hz,
                ZPK=bestfit_ZPK_s,
                F_nyquist_Hz=F_nyquist_Hz,
            )

        if delay_s is not None:
            data = data * np.exp(-2j * np.pi * delay_s * F_Hz)

    if bestfit_ZPK_z is not None:
        rep_z = representations.ZPKwData(
            data=data,
            F_Hz=F_Hz,
            W=SNR,
            F_nyquist_Hz=F_nyquist_Hz,
            ZPK=bestfit_ZPK_z,
            delay_s=delay_s,
        )
    else:
        rep_z = None

    if bestfit_ZPK_s is not None:
        rep_s = representations.ZPKwData(
            data=data,
            F_Hz=F_Hz,
            W=SNR,
            F_nyquist_Hz=None,
            ZPK=bestfit_ZPK_s,
            delay_s=delay_s,
        )
    else:
        rep_s = None

    return Bunch(
        F_Hz=F_Hz,
        data=data,
        SNR=SNR,
        F_nyquist_Hz=F_nyquist_Hz,
        residuals_log_best=residuals_log_best,
        bestfit_ZPK_z=bestfit_ZPK_z,
        bestfit_ZPK_s=bestfit_ZPK_s,
        rep_z=rep_z,
        rep_s=rep_s,
        iN=iN,
        sN=sN,
        **kwargs
    )


def assert_almost_equal(arr1, arr2, decimals):
    np.testing.assert_allclose(arr1, arr2, decimals)
    # np.testing.assert_allclose(arr1.real, arr2.real, decimals)
    # np.testing.assert_allclose(arr1.imag, arr2.imag, decimals)


def sign_validate(aid, fitter):
    """
    To be added as a hint to data2filter
    """
    rep = fitter.ZPKrep
    xfer = rep.xfer_fit
    data = rep.data
    rat = data / xfer
    rat_ang = np.exp(1j * np.angle(rat))
    ang_avg_rep = np.sum(rat_ang * rep.W ** 2) / np.sum(rep.W ** 2)

    xfer = fitter.xfer_fit
    data = fitter.data
    rat = data / xfer
    rat_ang = np.exp(1j * np.angle(rat))
    ang_avg_fit = np.sum(rat_ang * fitter.W ** 2) / np.sum(fitter.W ** 2)
    # print("SGN: ", ang_avg_rep, ang_avg_fit)
    # axB = plots.plots.plot_fit(
    #    fitter,
    #    fname = 'test1.png',
    # )
    # axB = plots.plots.plot_fitter_flag(
    #    fitter,
    #    fname = 'test3.png',
    # )
    # axB = plots.plots.plot_fit(
    #    fitter.ZPKrep,
    #    fname = 'test2.png',
    # )

    if isinstance(fitter, fitters_ZPK.MultiReprFilterBase):
        for coding in list(fitter.num_codings) + list(fitter.den_codings):
            rB = representations.RootBunch(
                u=coding.roots(),
                constraint=representations.root_constraints.no_constraint,
            )
            h1 = coding.transfer()
            h, lnG = rB.val_lnG(fitter.Xex_grid)
            h = h * np.exp(lnG)
            assert_almost_equal(h / h1, 1, 4)

    assert ang_avg_fit.real > 0 and ang_avg_rep.real > 0


sign_validate_hint = {
    "fitter_update_validate": sign_validate,
    "fitter_check_validate": sign_validate,
}


def rational_fitter_validate(rat_fitter, fitter):
    """
    To be added as a hint to data2filter
    """
    rep1 = rat_fitter.ZPKrep.xfer_fit
    rep2 = fitter.ZPKrep.xfer_fit
    rat = rat_fitter.xfer_fit
    mrf = fitter.xfer_fit

    assert_almost_equal(rep1 / rep2, 1, 5)
    assert_almost_equal(rat / rep2, 1, 5)
    assert_almost_equal(mrf / rep2, 1, 5)
    print("Checking Rational Fitter")


def sign_validate_and_plot_hint(pyfile, request):
    def sign_validate_plot(aid, fitter):
        with plots.plot_on_assert(pyfile, request, fitter, plot_anyway=False):
            try:
                sign_validate(aid, fitter)
            except AssertionError:
                print(fitter)
                print(fitter.F_nyquist_Hz)
                print(fitter.zeros)
                print(fitter.zeros_overlay)
                print(fitter.ZPKrep)
                # assert(False)
                raise

    hint = {
        "fitter_update_validate": sign_validate_plot,
        "fitter_check_validate": sign_validate_plot,
        "rational_fitter_validate": rational_fitter_validate,
    }
    return hint


def stability_validate_and_plot_hint(pyfile, request):
    def sign_validate_plot(aid, fitter):
        with plots.plot_on_assert(pyfile, request, fitter, plot_anyway=False):
            assert np.all(fitter.ZPKrep.poles.c.real <= 0)
            assert np.all(fitter.ZPKrep.poles.r.real <= 0)
            try:
                sign_validate(aid, fitter)
            except AssertionError:
                print(fitter)
                print(fitter.F_nyquist_Hz)
                print(fitter.zeros)
                print(fitter.zeros_overlay)
                print(fitter.ZPKrep)
                # assert(False)
                raise

    hint = {
        "fitter_update_validate": sign_validate_plot,
        "fitter_check_validate": sign_validate_plot,
        "rational_fitter_validate": rational_fitter_validate,
    }
    return hint


def sign_validate_and_digest_hint(pyfile, request):
    def sign_validate_plot(aid, fitter):
        with plots.digest_on_assert(pyfile, request, aid, plot_anyway=False):
            try:
                sign_validate(aid, fitter)
            except AssertionError:
                print(fitter)
                print(fitter.F_nyquist_Hz)
                print(fitter.zeros)
                print(fitter.zeros_overlay)
                print(fitter.ZPKrep)
                # assert(False)
                raise

    hint = {
        "fitter_update_validate": sign_validate_plot,
        "fitter_check_validate": sign_validate_plot,
        "rational_fitter_validate": rational_fitter_validate,
    }
    return hint



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

from wield.iirrational.testing import IIRrational_data
from wield.iirrational import plots

# from wield.iirrational.utilities.mpl import mplfigB
import os.path as path

from wield.iirrational.fitters_ZPK import ZPKrep2MRF
from wield.iirrational.fitters_ZPK import codings_z
from wield.iirrational import testing

from utilities import check_residuals


@pytest.mark.parametrize("coding_map", sorted(codings_z.coding_maps.keys()))
def test_NL_fitter(coding_map):
    """
    tests when the fitter is bad as it is trying to fit on the wrong side
    """
    data = IIRrational_data("testZPK4x")
    # data.rep_z.zeros = data.rep_z.zeros
    # data.rep_z.poles = data.rep_z.poles
    fitterMRF = ZPKrep2MRF(
        data.rep_z.ZPKrep,
        coding_map=codings_z.coding_maps[coding_map],
    )
    testing.assert_almost_equal(data.rep_z.xfer_fit / fitterMRF.xfer_fit, 1, 4)
    testing.assert_almost_equal(data.rep_z.ZPKrep.xfer_fit / fitterMRF.xfer_fit, 1, 4)

    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    opt = fitterMRF.optimize()
    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    # axB = plot_fitter_flag(fitterMRF, with_error=False, plot_zp = True)


@pytest.mark.parametrize("coding_map", sorted(codings_z.coding_maps.keys()))
def test_NL_fitter_US(coding_map):
    data = IIRrational_data("testZPK4x")
    data.rep_z.zeros = 1.01 * (data.rep_z.zeros)
    data.rep_z.poles = 0.99 * (data.rep_z.poles)
    fitterMRF = ZPKrep2MRF(
        data.rep_z.ZPKrep,
        coding_map=codings_z.coding_maps[coding_map],
    )

    testing.assert_almost_equal(data.rep_z.xfer_fit / fitterMRF.xfer_fit, 1, 4)
    testing.assert_almost_equal(data.rep_z.ZPKrep.xfer_fit / fitterMRF.xfer_fit, 1, 4)
    print(data.rep_z.F_nyquist_Hz)
    print(fitterMRF.F_nyquist_Hz)

    try:
        print(fitterMRF.residuals_average)
        check_residuals(fitterMRF)
        print(fitterMRF.zeros)
        opt = fitterMRF.optimize()
        print(fitterMRF.zeros)
        print(fitterMRF.residuals_average)
        # the check after optimizing to unstable roots is... bad
        # check_residuals(fitterMRF)
        # assert(fitterMRF.residuals_average < 10)
    except AssertionError:
        # TODO make this plot_on_assert
        axB = plots.plot_fitter_flag(fitterMRF, with_error=False, plot_zp=True)
        axB.save(path.join(path.split(__file__)[0], "out-Z/{}.png".format(coding_map)))
        raise


@pytest.mark.parametrize("coding_map", sorted(codings_z.coding_maps.keys()))
def test_NL_fitter1_DL(coding_map):
    """
    tests when the fitter is bad as it is trying to fit on the wrong side
    """
    # set number 1 is logarithmic data
    data = IIRrational_data("testZPK4x", set_num=1)
    data.rep_z.zeros = data.rep_z.zeros
    data.rep_z.poles = data.rep_z.poles
    print(data.rep_z.zeros)
    fitterMRF = ZPKrep2MRF(
        data.rep_z.ZPKrep,
        coding_map=codings_z.coding_maps[coding_map],
        distance_limit_auto=1,
    )
    testing.assert_almost_equal(data.rep_z.xfer_fit / fitterMRF.xfer_fit, 1, 2)
    testing.assert_almost_equal(data.rep_z.ZPKrep.xfer_fit / fitterMRF.xfer_fit, 1, 2)
    fitterMRF.distance_limit_auto = 1

    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    opt = fitterMRF.optimize()
    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    # axB = plot_fitter_flag(fitterMRF, with_error=False, plot_zp = True)


@pytest.mark.parametrize("coding_map", sorted(codings_z.coding_maps.keys()))
def test_NL_fitter1_DL2(coding_map):
    """
    tests when the fitter is bad as it is trying to fit on the wrong side
    """
    # set number 1 is logarithmic data
    data = IIRrational_data("testZPK4x", set_num=1)
    data.rep_z.zeros = data.rep_z.zeros
    data.rep_z.poles = data.rep_z.poles
    print(data.rep_z.zeros)
    fitterMRF = ZPKrep2MRF(
        data.rep_z.ZPKrep,
        coding_map=codings_z.coding_maps[coding_map],
        distance_limit_auto=1,
    )
    testing.assert_almost_equal(data.rep_z.xfer_fit / fitterMRF.xfer_fit, 1, 2)
    testing.assert_almost_equal(data.rep_z.ZPKrep.xfer_fit / fitterMRF.xfer_fit, 1, 2)
    fitterMRF.distance_limit_auto = 2

    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    opt = fitterMRF.optimize()
    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    # axB = plot_fitter_flag(fitterMRF, with_error=False, plot_zp = True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np
import scipy.signal


def SNR_estimate(F_Hz, xfer, width=5):
    # xfer = xfer / abs(xfer)

    xferL = np.log(abs(xfer))
    d = np.cumsum(xferL)
    diff = (d[width:] - d[:-width]) / width
    diff = np.concatenate(
        [
            diff,
            d[-1] - d[-width - 1 : -1] / np.arange(1 + width, 1, -1),
        ]
    )
    R1 = xferL - diff
    R1d = R1[1:] - R1[:-1]

    xferP = xfer / abs(xfer)
    d = np.cumsum(xferP)
    diff = (d[width:] - d[:-width]) / width
    diff = np.concatenate(
        [
            diff,
            d[-1] - d[-width - 1 : -1] / np.arange(1 + width, 1, -1),
        ]
    )
    R2 = xferP / diff
    R2d = R2[1:] - R2[:-1]

    W_c = np.cumsum(np.maximum(abs(R2d) ** 2, abs(R1d) ** 2))
    W_diff = (W_c[width:] - W_c[:-width]) / width
    W_est = np.maximum(1 / W_diff - 0.5, 0) ** 0.5
    W_est = np.concatenate(
        [
            W_est,
            [W_est[-1]] * (width + 1),
        ]
    )

    return W_est



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import sys

from wield.bunch.depbunch import (
    depB_property,
    DepBunch,
)
import numpy as np
from ... import TFmath


def abs_sq(x):
    return x.real ** 2 + x.imag ** 2


# root_constraints:
class RootConstraints(object):
    no_constraint = frozenset()  # real Even real Odd
    mirror_real = frozenset(["mirror_real"])  # real Even real Odd
    mirror_imag = frozenset(["mirror_imag"])  # real Even imaginary Odd
    mirror_disc = frozenset(["mirror_disc"])  # palendromic standard polynomials

    mirror_quad = mirror_real | mirror_imag

    mirror_real_disc = mirror_real | mirror_disc
    mirror_imag_disc = mirror_imag | mirror_disc
    mirror_quad_disc = mirror_quad | mirror_disc


root_constraints = RootConstraints()

# rootBunches must have the following elements
# rB.constraint which defines how to interpret
# rB.z is the number of zero roots
# rB.c are complex roots
# rB.r are real roots
# rB.i are imaginary roots
# rB.dc are disc complex roots
# rB.dr are disc real roots
# rB.di are disc imaginary roots
# rB.u are unsorted roots (not allowed for many applications)

# if the mirror_imag is not specified, imaginary roots will not be separated
# if the mirror_real is not specified, real roots will not be separated
# if the mirror_disc is not specified, disc roots will not be separated

# if mirror flags are given, only roots non-redundant are provided in the lists
# zero roots will always be in the real list
#


class RootBunch(DepBunch):
    root_constraints = root_constraints

    def __init__(
        self,
        constraint=None,
        z=None,
        c=None,
        r=None,
        i=None,
        dc=None,
        dr=None,
        di=None,
        u=(),
        copy=None,
        **kwargs
    ):
        if copy is not None:
            # TODO, make this more concise
            return super(RootBunch, self).__init__(
                self,
                constraint=constraint,
                z=z,
                c=i,
                r=r,
                i=i,
                dc=dc,
                dr=dr,
                di=di,
                u=u,
                copy=copy,
                **kwargs
            )
        remainder = constraint - root_constraints.mirror_quad_disc
        if remainder:
            raise RuntimeError("Unknown root_constraints: {0}".format(remainder))
        if copy is not None:
            super(RootBunch, self).__init__(copy=copy, **kwargs)
            return

        if constraint is None:
            raise TypeError("Must specify constraint argument")

        super(RootBunch, self).__init__(**kwargs)
        self.constraint = constraint
        if z is not None:
            self.z = int(z)
        if c is not None:
            self.c = np.asarray(c)
        if r is not None:
            # TODO, check realness
            self.r = np.asarray(r)
        if i is not None:
            # TODO, check imaginaryness
            self.i = np.asarray(i)
        if dc is not None:
            self.dc = np.asarray(dc)
        if dr is not None:
            # TODO, check realness
            self.dr = np.asarray(dr)
        if di is not None:
            # TODO, check imaginaryness
            self.di = np.asarray(di)
        self.u = np.asarray(u)

        if root_constraints.mirror_disc <= constraint:
            assert dc is not None
            if root_constraints.mirror_quad <= constraint:
                assert z is not None
                assert c is not None
                assert r is not None
                assert i is not None
                assert dc is not None
                assert dr is not None
                assert di is not None
                # TODO, needs even more asserts
                assert np.all(self.r.real < 0)
                assert np.all(self.i.real > 0)
                assert np.all(self.r.imag == 0)
                assert np.all(self.i.imag == 0)
                assert np.all(self.dr.real < 0)
                assert np.all(self.di.real > 0)
                assert np.all(self.dr.imag == 0)
                assert np.all(self.di.imag == 0)
            elif root_constraints.mirror_real <= constraint:
                assert z is None
                assert r is not None
                assert c is not None
                assert dr is not None
                assert dc is not None
                # TODO, needs even more asserts
                assert np.all(self.c.imag > 0)
                assert np.all(self.dc.imag > 0)
                assert np.all(self.r.imag == 0)
                assert np.all(self.dr.imag == 0)
            elif root_constraints.mirror_imag <= constraint:
                assert z is None
                assert i is not None
                assert c is not None
                assert di is not None
                assert dc is not None
                # TODO, needs even more asserts
                assert np.all(self.c.real > 0)
                assert np.all(self.dc.real > 0)
                assert np.all(self.i.imag == 0)
                assert np.all(self.di.imag == 0)
        else:
            assert dc is None
            assert dr is None
            assert di is None
            if root_constraints.mirror_quad <= constraint:
                assert z is not None
                assert c is not None
                assert r is not None
                assert i is not None
                assert np.all(self.c.real < 0)
                assert np.all(self.c.imag > 0)
                assert np.all(self.r.real < 0)
                assert np.all(self.i.real > 0)
                assert np.all(self.r.imag == 0)
                assert np.all(self.i.imag == 0)
            elif root_constraints.mirror_real <= constraint:
                assert z is None
                assert r is not None
                assert c is not None
                assert i is None
                assert np.all(self.c.imag > 0)
                assert np.all(self.r.imag == 0)
            elif root_constraints.mirror_imag <= constraint:
                assert z is None
                assert i is not None
                assert c is not None
                assert r is None
                assert np.all(self.c.real > 0)
                assert np.all(self.i.imag == 0)
            else:
                assert z is None
                assert i is None
                assert c is None
                assert r is None
        return

    def __len__(self):
        return self.length

    def clear(self):
        self._count = self._count + 1

    def root_str_list(self):
        slist = []
        if root_constraints.mirror_disc <= self.constraint:
            raise NotImplementedError()
            if root_constraints.mirror_quad <= self.constraint:
                pass
            elif root_constraints.mirror_real <= self.constraint:
                pass
            elif root_constraints.mirror_imag <= self.constraint:
                pass
            else:
                # palidromic but complex
                pass
        else:
            if root_constraints.mirror_quad <= self.constraint:
                for r in sorted(self.r):
                    slist.append("±{}".format(r))
                for i in sorted(self.i):
                    slist.append("±{}j".format(i))
                for c in sorted(self.c, key=lambda v: v.imag):
                    slist.append("±{}±{}j".format(c.real, c.imag))
            elif root_constraints.mirror_real <= self.constraint:
                for r in sorted(self.r):
                    slist.append("{}".format(r))
                for c in sorted(self.c, key=lambda v: v.imag):
                    slist.append("{}±{}j".format(c.real, c.imag))
            elif root_constraints.mirror_imag <= self.constraint:
                for i in sorted(self.i):
                    slist.append("{}j".format(i))
                for c in sorted(self.c, key=lambda v: v.imag):
                    if c.imag == 0:
                        slist.append("±{}".format(c.real))
                    elif c.imag > 0:
                        slist.append("±{}+{}j".format(c.real, c.imag))
                    else:
                        slist.append("±{}-{}j".format(c.real, -c.imag))
            else:
                pass
        for u in sorted(self.u, key=lambda v: v.real):
            if u.imag == 0:
                slist.append("{}".format(u.real))
            elif u.imag > 0:
                slist.append("{}+{}j".format(u.real, u.imag))
            else:
                slist.append("{}-{}j".format(u.real, -u.imag))
        return slist

    def __str__(self):
        slist = self.root_str_list()
        # figure out how to make this pretty print with lines!

        s = "rB({})".format(", ".join(slist))
        # TODO, make this nicer
        if sys.version_info < (3, 0):
            s = s.encode("utf-8")
        return s

    # TODO make a REPR

    @depB_property
    def _count(self, val=0):
        return val

    @depB_property
    def length(self):
        self.dependencies("length")
        if root_constraints.mirror_disc <= self.constraint:
            raise NotImplementedError()
            if root_constraints.mirror_quad <= self.constraint:
                pass
            elif root_constraints.mirror_real <= self.constraint:
                pass
            elif root_constraints.mirror_imag <= self.constraint:
                pass
            else:
                # palidromic but complex
                pass
        else:
            if root_constraints.mirror_quad <= self.constraint:
                return (
                    self.z
                    + 2 * self.r.size
                    + 2 * self.i.size
                    + 4 * self.c.size
                    + self.u.size
                )
            elif root_constraints.mirror_real <= self.constraint:
                return self.r.size + 2 * self.c.size + self.u.size
            elif root_constraints.mirror_imag <= self.constraint:
                return self.i.size + 2 * self.c.size + self.u.size
            else:
                return len(self.u)

    @depB_property
    def fullplane(self):
        self.dependencies("length")
        if root_constraints.mirror_disc <= self.constraint:
            raise NotImplementedError()
            if root_constraints.mirror_quad <= self.constraint:
                pass
            elif root_constraints.mirror_real <= self.constraint:
                pass
            elif root_constraints.mirror_imag <= self.constraint:
                pass
            else:
                # palidromic but complex
                pass
        else:
            if root_constraints.mirror_quad <= self.constraint:
                a = np.empty(
                    (
                        self.z
                        + 2 * self.r.size
                        + 2 * self.i.size
                        + 4 * self.c.size
                        + self.u.size,
                    ),
                    dtype=self.c.dtype,
                )
                a[: self.z] = 0
                offset = self.z
                offset_2 = offset + 2 * self.r.size
                v = a[offset:offset_2]
                v[::2] = self.r
                v[1::2] = -self.r
                offset = offset_2
                offset_2 = offset + 2 * self.i.size
                v = a[offset:offset_2]
                v[::2] = 1j * self.i
                v[1::2] = -1j * self.i
                offset = offset_2
                offset_2 = offset + 4 * self.c.size
                v = a[offset:offset_2]
                v[::4] = self.c
                v[1::4] = self.c.conjugate()
                v[2::4] = -self.c
                v[3::4] = -self.c.conjugate()
                return a
                offset = offset_2
                offset_2 = offset + self.u.size
                v = a[offset:offset_2]
                v[:] = self.u
            elif root_constraints.mirror_real <= self.constraint:
                a = np.empty(
                    (self.r.size + 2 * self.c.size + self.u.size), dtype=self.c.dtype
                )
                offset = 0
                offset_2 = offset + 1 * self.r.size
                v = a[offset:offset_2]
                v[:] = self.r
                offset = offset_2
                offset_2 = offset + 2 * self.c.size
                v = a[offset:offset_2]
                v[::2] = self.c
                v[1::2] = self.c.conjugate()
                offset = offset_2
                offset_2 = offset + self.u.size
                v = a[offset:offset_2]
                v[:] = self.u
                return a
            elif root_constraints.mirror_imag <= self.constraint:
                a = np.empty(
                    (self.i.size + 2 * self.c.size + self.u.size), dtype=self.c.dtype
                )
                offset = 0
                offset_2 = offset + self.i.size
                v = a[offset:offset_2]
                v[:] = 1j * self.i
                offset = offset_2
                offset_2 = offset + 2 * self.c.size
                v = a[offset:offset_2]
                v[::2] = self.c
                v[1::2] = -self.c
                offset = offset_2
                offset_2 = offset + self.u.size
                v = a[offset:offset_2]
                v[:] = self.u
                return a
            else:
                return self.u

    def multiply_by(self, scale_factor):
        if np.imag(scale_factor) == 0:
            if root_constraints.mirror_disc <= self.constraint:
                raise NotImplementedError()
            else:
                if root_constraints.mirror_quad == self.constraint:
                    return self.__class__(
                        constraint=self.constraint,
                        z=self.z,
                        u=scale_factor * self.u,
                        c=scale_factor * self.c,
                        r=scale_factor * self.r,
                        i=scale_factor * self.i,
                    )
                elif root_constraints.mirror_real == self.constraint:
                    return self.__class__(
                        constraint=self.constraint,
                        u=scale_factor * self.u,
                        c=scale_factor * self.c,
                        r=scale_factor * self.r,
                    )
                elif root_constraints.mirror_imag == self.constraint:
                    return self.__class__(
                        constraint=self.constraint,
                        u=scale_factor * self.u,
                        c=scale_factor * self.c,
                        i=scale_factor * self.i,
                    )
                elif root_constraints.no_constraint == self.constraint:
                    return self.__class__(
                        constraint=self.constraint,
                        u=scale_factor * self.u,
                    )
                else:
                    raise RuntimeError("Unrecognized Root_constraints")

        elif np.real(scale_factor) == 0:
            scale_factor_i = scale_factor.imag
            if root_constraints.mirror_disc <= self.constraint:
                raise NotImplementedError()
            else:
                # TODO
                if root_constraints.mirror_quad == self.constraint:
                    scale_factor_abs = abs(scale_factor_i)
                    return self.__class__(
                        constraint=root_constraints.mirror_quad,
                        z=self.z,
                        u=1j * scale_factor_i * self.u,
                        c=(1j * scale_factor_abs * self.c).conjugate(),
                        r=-scale_factor_abs * self.i,
                        i=-scale_factor_abs * self.r,
                    )
                elif root_constraints.mirror_real == self.constraint:
                    if scale_factor_i > 0:
                        return self.__class__(
                            constraint=root_constraints.mirror_imag,
                            u=scale_factor_i * self.u,
                            c=(1j * scale_factor_i * self.c.conjugate()),
                            i=scale_factor_i * self.r,
                        )
                    else:
                        return self.__class__(
                            constraint=root_constraints.mirror_imag,
                            u=scale_factor_i * self.u,
                            c=(1j * scale_factor_i * self.c),
                            i=scale_factor_i * self.r,
                        )
                elif root_constraints.mirror_imag == self.constraint:
                    if scale_factor_i > 0:
                        return self.__class__(
                            constraint=root_constraints.mirror_real,
                            u=scale_factor_i * self.u,
                            c=(1j * scale_factor_i * self.c),
                            r=-scale_factor_i * self.i,
                        )
                    else:
                        return self.__class__(
                            constraint=root_constraints.mirror_real,
                            u=scale_factor_i * self.u,
                            c=(1j * scale_factor_i * self.c.conjugate()),
                            r=-scale_factor_i * self.i,
                        )
                elif root_constraints.no_constraint == self.constraint:
                    return self.__class__(
                        constraint=self.constraint,
                        u=scale_factor_i * self.u,
                    )
                else:
                    raise RuntimeError("Unrecognized Root_constraints")
        else:
            raise NotImplementedError("Cannot rotate continuous angles")

    def _multiply_RB(self, other):
        """ """
        if other.constraint != self.constraint:
            raise RuntimeError("Can't multiply root bunches with differing constraints")

        mR = bool(root_constraints.mirror_real & self.constraint)
        mI = bool(root_constraints.mirror_imag & self.constraint)
        mD = bool(root_constraints.mirror_disc & self.constraint)
        kw = dict()
        kw["u"] = np.concatenate([self.u, other.u])
        if mR or mI:
            kw["c"] = np.concatenate([self.c, other.c])
            if mR:
                kw["r"] = np.concatenate([self.r, other.r])
                if mI:
                    kw["i"] = np.concatenate([self.i, other.i])
                    kw["z"] = self.z + other.z
                    if mD:
                        kw["di"] = np.concatenate([self.di, other.di])
                        kw["dr"] = np.concatenate([self.dr, other.dr])
                        kw["dc"] = np.concatenate([self.dc, other.dc])
                elif mD:
                    kw["dr"] = np.concatenate([self.dr, other.dr])
                    kw["dc"] = np.concatenate([self.dc, other.dc])
            else:
                # mR not true, mI true
                kw["i"] = np.concatenate([self.i, other.i])
                if mD:
                    kw["di"] = np.concatenate([self.di, other.di])
                    kw["dc"] = np.concatenate([self.dc, other.dc])
        return self.__class__(constraint=self.constraint, **kw)

    def __mul__(self, other):
        if isinstance(other, RootBunch):
            return self._multiply_RB(other)
        return self.multiply_by(other)

    def __rmul__(self, other):
        return self.multiply_by(other)

    def __truediv__(self, other):
        return self.multiply_by(1 / other)

    def __div__(self, other):
        return self.multiply_by(1 / other)

    def val_lnG(self, X, h=1, lnG=0):
        """
        returns the value as if it were generated from a polynomial with last coefficient 1 given a coefficient representation
        and the X_scale. It computes this without having to convert to the coefficient representation (how efficient!).
        """
        X = np.asarray(X)
        h = np.array(h, copy=True, dtype=np.complex128)
        X, h = np.broadcast_arrays(X, h)

        # note that this modifies in-place
        def VfR(roots, h, lnG):
            if len(roots) == 0:
                return h, lnG
            roots = np.asarray(roots)
            mlen = len(roots)
            group_len = 5
            for idx in range((mlen - 1) // group_len + 1):
                r = roots[idx * group_len : (idx + 1) * group_len]
                h = h * np.polynomial.polynomial.polyvalfromroots(X, r)
                abs_max = np.max(abs_sq(h)) ** 0.5
                h /= abs_max
                lnG += np.log(abs_max)
            return h, lnG

        if len(self.u) > 0:
            h, lnG = VfR(self.u, h, lnG)

        mR = bool(root_constraints.mirror_real & self.constraint)
        mI = bool(root_constraints.mirror_imag & self.constraint)
        mD = bool(root_constraints.mirror_disc & self.constraint)
        if mR or mI:
            h, lnG = VfR(self.c, h, lnG)
            if mR:
                h, lnG = VfR(self.c.conjugate(), h, lnG)
                h, lnG = VfR(self.r, h, lnG)
                if mI:
                    h, lnG = VfR(-self.c, h, lnG)
                    h, lnG = VfR(-self.c.conjugate(), h, lnG)
                    h, lnG = VfR(-self.r, h, lnG)
                    h, lnG = VfR(1j * self.i, h, lnG)
                    h, lnG = VfR(-1j * self.i, h, lnG)
                    if mD:
                        h, lnG = VfR(1 / self.c.conjugate(), h, lnG)
                        h, lnG = VfR(1 / self.c, h, lnG)
                        h, lnG = VfR(-1 / self.c.conjugate(), h, lnG)
                        h, lnG = VfR(-1 / self.c, h, lnG)
                        h, lnG = VfR(1 / self.r, h, lnG)
                        h, lnG = VfR(-1 / self.r, h, lnG)
                        h, lnG = VfR(1j / self.i, h, lnG)
                        h, lnG = VfR(-1j / self.i, h, lnG)
                elif mD:
                    h, lnG = VfR(1 / self.c.conjugate(), h, lnG)
                    h, lnG = VfR(1 / self.c, h, lnG)
                    h, lnG = VfR(1 / self.r, h, lnG)
            else:
                # mR not true, mI true
                h, lnG = VfR(-self.c, h, lnG)
                h, lnG = VfR(1j * self.i, h, lnG)
                if mD:
                    h, lnG = VfR(1 / self.c.conjugate(), h, lnG)
                    h, lnG = VfR(-1 / self.c.conjugate(), h, lnG)
                    h, lnG = VfR(1j / self.i, h, lnG)
        return h, lnG


class RBAlgorithms(object):
    root_constraints = root_constraints

    def __init__(
        self,
        strict=True,
        line_tol=1e-6,
        match_tol=1e-4,
        zero_tol=1e-10,
        lax_line_tol=0,
        constraint_standard=None,
    ):
        self.strict = strict
        self.lax_line_tol = lax_line_tol
        self.constraint_standard = constraint_standard

        def are_same(r1, r2):
            if abs(r1) < 0.8:
                return abs(r1 - r2) < match_tol
            else:
                return abs((r1 / r2) - 1) < match_tol

        def are_real(r1):
            if abs(r1.real) < 0.8:
                return abs(np.imag(r1)) < line_tol
            else:
                return abs(np.imag(r1) / np.real(r1)) < line_tol

        def are_imag(r1):
            if abs(r1.imag) < 0.8:
                return abs(np.real(r1)) < line_tol
            else:
                return abs(np.real(r1) / np.imag(r1)) < line_tol

        def are_zero(r1):
            return abs(np.imag(r1)) < zero_tol

        self.are_same = np.vectorize(are_same, otypes=[bool])
        self.are_real = np.vectorize(are_real, otypes=[bool])
        self.are_imag = np.vectorize(are_imag, otypes=[bool])
        self.are_zero = np.vectorize(are_zero, otypes=[bool])

    def expect_atleast(
        self,
        rB,
        constraint=None,
        allow_unknown=False,
    ):
        if constraint is None:
            constraint = self.constraint_standard
            # root_constraints.no_constraint
        if not isinstance(rB, RootBunch):
            rB = RootBunch(
                u=np.asarray(rB),
                constraint=root_constraints.no_constraint,
            )

        if constraint <= rB.constraint:
            return rB

        return self.expect(rB, constraint=constraint, allow_unknown=allow_unknown)

    def expect(
        self,
        rB,
        constraint=None,
        allow_unknown=False,
    ):
        if constraint is None:
            constraint = self.constraint_standard
        # real root_constraints.mirrors first
        # imaginary root_constraints.mirrors second
        # now do disc root_constraints.mirrors
        remainder = constraint - root_constraints.mirror_quad_disc
        if remainder:
            raise RuntimeError("Unknown root_constraints: {0}".format(remainder))

        if not isinstance(rB, RootBunch):
            rB = RootBunch(
                u=np.asarray(rB),
                constraint=root_constraints.no_constraint,
            )
            return self.expect(rB, constraint, allow_unknown=allow_unknown)

        elif rB.constraint == constraint:
            if constraint == root_constraints.no_constraint:
                return rB
            elif not allow_unknown and len(rB.u) > 0:
                raise RuntimeError("Roots contain unpaired roots: {}".format(rB.u))
            return rB
        if rB.constraint <= constraint:
            return self._expect_elevate(
                rB,
                constraint,
                allow_unknown=allow_unknown,
            )
        elif constraint <= rB.constraint:
            return self._expect_reduce(
                rB,
                constraint,
                allow_unknown=allow_unknown,
            )
        else:
            # reduce to common root_constraints
            rB = self._expect_reduce(
                rB,
                (constraint & rB.constraint),
                allow_unknown=allow_unknown,
            )
            # now elevate to the new constraint
            return self._expect_elevate(
                rB,
                constraint,
                allow_unknown=allow_unknown,
            )
        raise NotImplementedError("Can't Get here")
        return

    def _expect_elevate(self, rB, constraint, allow_unknown):
        # can only be called if rB.constraint <= constraint and rB.constraint != constraint
        assert rB.constraint != constraint

        def check_allow_unknown(rB):
            if not allow_unknown:
                if len(rB.u) > 0:
                    raise RuntimeError("Roots contain unpaired roots: {}".format(rB.u))

        def recurse(rB):
            check_allow_unknown(rB)
            if rB.constraint == constraint:
                return rB
            return self._expect_elevate(
                rB,
                constraint,
                allow_unknown=allow_unknown,
            )

        # elevate root_constraints
        if root_constraints.mirror_disc <= constraint:
            raise NotImplementedError(
                "Can't generate root_constraints.mirror_disc root_constraints yet"
            )
        elif root_constraints.mirror_real == rB.constraint:
            rB = self.MR2MQ(
                roots_c=rB.c,
                roots_r=rB.r,
                roots_u=rB.u,
            )
            return recurse(rB)
        elif root_constraints.mirror_imag == rB.constraint:
            rB = self.MI2MQ(
                roots_c=rB.c,
                roots_i=rB.i,
                roots_u=rB.u,
            )
            return recurse(rB)
        elif root_constraints.no_constraint == rB.constraint:
            if root_constraints.mirror_real <= constraint:
                rB = self.NC2MR(
                    roots_u=rB.u,
                )
                return recurse(rB)
            elif root_constraints.mirror_imag <= constraint:
                rB = self.NC2MI(
                    roots_u=rB.u,
                )
                return recurse(rB)
            else:
                raise NotImplementedError("Unknown final constraint")
        else:
            raise NotImplementedError("Unknown constraint in argument")
        return

    def _expect_reduce(self, rB, constraint, allow_unknown):
        def check_allow_unknown(rB):
            if constraint == root_constraints.no_constraint:
                pass
            elif not allow_unknown and len(rB.u) > 0:
                raise RuntimeError("Roots contain unpaired roots")

        def recurse(rB):
            check_allow_unknown(rB)
            return self._expect_reduce(
                rB,
                constraint,
                allow_unknown=allow_unknown,
            )

        if constraint == root_constraints.no_constraint:
            return RootBunch(
                u=rB.fullplane,
                constraint=root_constraints.no_constraint,
            )
        # reduce root_constraints
        raise NotImplementedError("Can not currently reduce root_constraints")
        if root_constraints.mirror_disc <= rB.constraint:
            raise NotImplementedError(
                "Can't handle root_constraints.mirror disc root_constraints"
            )
        elif root_constraints.mirror_quad <= rB.constraint:
            if root_constraints.mirror_real <= constraint:
                raise NotImplementedError()
                return
            else:
                # must be root_constraints.mirror_imag <= constraint
                raise NotImplementedError()
                return
            return
        elif root_constraints.mirror_real <= rB.constraint:
            raise NotImplementedError()
            return
        elif root_constraints.mirror_imag <= rB.constraint:
            raise NotImplementedError()
            return
        else:
            raise RuntimeError("Bad Logic")

    def NC2MR(self, roots_u):
        real_select = self.are_real(roots_u)
        roots_r = roots_u[real_select].real
        roots_u = roots_u[~real_select]

        pos_select = roots_u.imag > 0
        roots_c_neg = roots_u[~pos_select]
        roots_c_pos = roots_u[pos_select]
        rPB = TFmath.nearest_pairs(roots_c_pos, roots_c_neg.conjugate())
        if self.lax_line_tol > 0:
            roots_u = []
            roots_r2 = []

            def check_ins(u):
                if abs(u.real) > self.lax_line_tol:
                    if abs(u.imag) < self.lax_line_tol:
                        roots_r2.append(u.real)
                    else:
                        roots_u.append(u)
                else:
                    if abs(u.imag / u.real) < self.lax_line_tol:
                        roots_r2.append(u.real)
                    else:
                        roots_u.append(u)

            for u in rPB.l1_remain:
                check_ins(u)
            for u in rPB.l2_remain:
                check_ins(u.conjugate())
            roots_r = np.concatenate([roots_r, roots_r2])
        else:
            roots_u = list(rPB.l1_remain) + [r.conjugate() for r in rPB.l2_remain]
        roots_c = []
        for r1, r2 in rPB.r12_list:
            if not self.strict or self.are_same(r1, r2):
                # roots_c.append((r1 + r2) / 2)
                # TODO, this seems to work better, not clear why..
                roots_c.append(r1)
            else:
                roots_u.append(r1)
                roots_u.append(r2.conjugate())
        roots_c = np.array(roots_c)
        roots_u = np.array(roots_u)
        return RootBunch(
            constraint=root_constraints.mirror_real,
            c=roots_c,
            r=roots_r,
            u=roots_u,
        )

    def NC2MI(self, roots_u):
        imag_select = self.are_imag(roots_u)
        roots_i = roots_u[imag_select].imag
        roots_u = roots_u[~imag_select]

        pos_select = roots_u.real > 0
        roots_c_neg = roots_u[~pos_select]
        roots_c_pos = roots_u[pos_select]
        rPB = TFmath.nearest_pairs(roots_c_pos, -roots_c_neg.conjugate())

        # TODO, put this logic in the NC2MR

        # fill the list with as many pairs as possible
        r12_list_full = rPB.r12_list
        while rPB.r12_list:
            rPB = TFmath.nearest_pairs(rPB.l1_remain, rPB.l2_remain)
            r12_list_full.extend(rPB.r12_list)
        rPB.r12_list = r12_list_full

        if self.lax_line_tol > 0:
            roots_u = []
            roots_i2 = []

            def check_ins(u):
                if abs(u.imag) > self.lax_line_tol:
                    if abs(u.real) < self.lax_line_tol:
                        roots_i2.append(u.imag)
                    else:
                        roots_u.append(u)
                else:
                    if abs(u.real / u.imag) < self.lax_line_tol:
                        roots_i2.append(u.imag)
                    else:
                        roots_u.append(u)

            for u in rPB.l1_remain:
                check_ins(u)
            for u in rPB.l2_remain:
                check_ins(-u.conjugate())
            roots_i = np.concatenate([roots_i, roots_i2])
        else:
            roots_u = list(rPB.l1_remain) + [-r.conjugate() for r in rPB.l2_remain]

        roots_c = []
        for r1, r2 in rPB.r12_list:
            if not self.strict or self.are_same(r1, r2):
                # roots_c.append((r1 + r2) / 2)
                # TODO, this seems to work better, not clear why..
                roots_c.append(r1)
            else:
                roots_u.append(r1)
                roots_u.append(-r2.conjugate())
        roots_c = np.array(roots_c)
        roots_u = np.array(roots_u)
        return RootBunch(
            constraint=root_constraints.mirror_imag,
            c=roots_c,
            i=roots_i,
            u=roots_u,
        )

    def MI2MQ(self, roots_c, roots_i, roots_u):
        real_select = self.are_real(roots_c)
        roots_r = roots_c[real_select].real
        roots_c = roots_c[~real_select]

        pos_select = roots_c.imag > 0
        roots_c_neg = roots_c[~pos_select]
        roots_c_pos = roots_c[pos_select]
        rPB = TFmath.nearest_pairs(roots_c_pos, roots_c_neg.conjugate())
        roots_u2 = list(rPB.l1_remain) + [r.conjugate() for r in rPB.l2_remain]
        roots_c = []
        for r1, r2 in rPB.r12_list:
            if not self.strict or self.are_same(r1, r2):
                # roots_c.append((r1 + r2) / 2)
                # TODO, this seems to work better, not clear why..
                roots_c.append(r1)
            else:
                roots_u2.append(r1)
                roots_u2.append(r2.conjugate())
        roots_c = np.array(roots_c)
        roots_u = np.concatenate([roots_u, roots_u2])

        select_zero = self.are_zero(roots_i)
        roots_i = roots_i[~select_zero]
        z = np.count_nonzero(select_zero)
        return RootBunch(
            constraint=root_constraints.mirror_quad,
            z=z,
            c=roots_c,
            r=roots_r,
            i=roots_i,
            u=roots_u,
        )

    def MR2MQ(self, roots_c, roots_r, roots_u):
        imag_select = self.are_imag(roots_c)
        roots_i = roots_c[imag_select].imag
        roots_c = roots_c[~imag_select]

        pos_select = roots_c.real > 0
        roots_c_neg = roots_c[~pos_select]
        roots_c_pos = roots_c[pos_select]
        rPB = TFmath.nearest_pairs(roots_c_pos, -roots_c_neg.conjugate())
        roots_u2 = list(rPB.l1_remain) + [-r.conjugate() for r in rPB.l2_remain]
        roots_c = []
        for r1, r2 in rPB.r12_list:
            if not self.strict or self.are_same(r1, r2):
                # roots_c.append((r1 + r2) / 2)
                # TODO, this seems to work better, not clear why..
                roots_c.append(r1)
            else:
                roots_u2.append(r1)
                roots_u2.append(-r2.conjugate())
        roots_c = np.array(roots_c)
        roots_u = np.concatenate([roots_u, roots_u2])

        select_zero = self.are_zero(roots_r)
        roots_r = roots_r[~select_zero]
        z = np.count_nonzero(select_zero)
        return RootBunch(
            constraint=root_constraints.mirror_quad,
            z=z,
            c=roots_c,
            r=roots_r,
            i=roots_i,
            u=roots_u,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np

from ... import TFmath
from ...utilities import ensure_aid

from . import algorithms


def Q_rank_calc(z, p):
    if p.real == 0 or z.real == 0:
        if p.real == z.real:
            Q_rank = 0
        else:
            # TODO
            # should use the data spacing to regularize this case
            Q_rank = 1e3
    else:
        Q_rank = abs(p - z) * (1 / (p.real) ** 2 + 1 / (z.real) ** 2) ** 0.5
    return Q_rank


def Q_rank_single(r):
    Q_rank = abs(r.imag / r.real)
    return Q_rank


def ranking_reduction_cc(
    aid,
    marginalize_delay=True,
):
    aid = ensure_aid(aid)
    rank_zp_idx_list = []

    Pc = aid.fitter.poles.c
    Zc = aid.fitter.zeros.c

    # can always works since this is a balanced reduction

    # print(len(Zc), len(Pc))
    pairs = set()
    lZP = TFmath.nearest_idx(Zc, Pc)
    pairs.update(enumerate(lZP))

    lPZ = TFmath.nearest_idx(Pc, Zc)
    for idx_p, idx_z in enumerate(lPZ):
        pairs.add((idx_z, idx_p))

    lPZ = TFmath.nearest_idx(Pc, -Zc.conjugate())
    # only include ones which are pairing unstable with stable
    for idx_p, idx_z in enumerate(lPZ):
        if Pc[idx_p].real * Zc[idx_z].real < 0:
            pairs.add((idx_z, idx_p))

    if True:
        lPZ = TFmath.nearest_idx(Pc, Zc.conjugate())
        # include ones which are pairing unstable with stable
        for idx_p, idx_z in enumerate(lPZ):
            if Pc[idx_p].real * Zc[idx_z].real > 0:
                pairs.add((idx_z, idx_p))

    for idx_z, idx_p in pairs:
        if idx_z is None or idx_p is None:
            continue
        algorithms.resrank_program(
            aid.fitter,
            rank_zp_idx_list,
            "cc",
            ["ZcDelIdx", idx_z, "PcDelIdx", idx_p],
            marginalize_delay=marginalize_delay,
        )
    return rank_zp_idx_list


def ranking_reduction_c(
    aid,
    marginalize_delay=True,
):
    aid = ensure_aid(aid)
    rank_zp_idx_list = []

    Pc = aid.fitter.poles.c
    Zc = aid.fitter.zeros.c

    orders = aid.fitter_orders()
    rdmax = aid.hint("relative_degree_max")
    rdmin = aid.hint("relative_degree_min")

    if rdmin is None or orders.reldeg - 2 >= rdmin:
        for idx_z, z in enumerate(Zc):
            algorithms.resrank_program(
                aid.fitter,
                rank_zp_idx_list,
                "c",
                ["ZcDelIdx", idx_z],
                marginalize_delay=marginalize_delay,
            )

    if rdmax is None or orders.reldeg + 2 <= rdmax:
        for idx_p, p in enumerate(Pc):
            algorithms.resrank_program(
                aid.fitter,
                rank_zp_idx_list,
                "c",
                ["PcDelIdx", idx_p],
                marginalize_delay=marginalize_delay,
            )
    return rank_zp_idx_list


def ranking_reduction_r(
    aid,
    marginalize_delay=True,
):
    aid = ensure_aid(aid)
    rank_zp_idx_list = []

    Pr = aid.fitter.poles.r
    Zr = aid.fitter.zeros.r

    orders = aid.fitter_orders()
    rdmax = aid.hint("relative_degree_max")
    rdmin = aid.hint("relative_degree_min")

    if rdmin is None or orders.reldeg - 1 >= rdmin:
        for idx_z, z in enumerate(Zr):
            algorithms.resrank_program(
                aid.fitter,
                rank_zp_idx_list,
                "r",
                ["ZrDelIdx", idx_z],
                marginalize_delay=marginalize_delay,
            )

    if rdmax is None or orders.reldeg + 1 <= rdmax:
        for idx_p, p in enumerate(Pr):
            algorithms.resrank_program(
                aid.fitter,
                rank_zp_idx_list,
                "r",
                ["PrDelIdx", idx_p],
                marginalize_delay=marginalize_delay,
            )
    return rank_zp_idx_list


def ranking_reduction_r_flip(
    aid,
    marginalize_delay=True,
):
    aid = ensure_aid(aid)
    rank_zp_idx_list = []

    Zr = aid.fitter.zeros.r

    for idx_z, z in enumerate(Zr):
        z_flip = -z.conjugate()
        algorithms.resrank_program(
            aid.fitter,
            rank_zp_idx_list,
            "r",
            ["ZrDelIdx", idx_z, "ZrAdd", z_flip],
            marginalize_delay=marginalize_delay,
        )

    return rank_zp_idx_list


def ranking_reduction_rr(
    aid,
    marginalize_delay=True,
):
    aid = ensure_aid(aid)
    rank_zp_idx_list = []

    Pr = aid.fitter.poles.r
    Zr = aid.fitter.zeros.r

    # doesn't need to check reldeg since this preserves it

    pairs = set()
    lZP = TFmath.nearest_idx(Zr, Pr)
    pairs.update(enumerate(lZP))

    lPZ = TFmath.nearest_idx(Pr, Zr)
    for idx_p, idx_z in enumerate(lPZ):
        pairs.add((idx_z, idx_p))

    # TODO, include these via an option in case of starvation?
    lPZ = TFmath.nearest_idx(Pr, -Zr)
    # only include ones which are pairing unstable with stable
    for idx_p, idx_z in enumerate(lPZ):
        if Pr[idx_p] * Zr[idx_z] < 0:
            pairs.add((idx_z, idx_p))

    if True:
        lPZ = TFmath.nearest_idx(Pr, Zr)
        # include ones which are pairing stable with stable
        for idx_p, idx_z in enumerate(lPZ):
            if Pr[idx_p] * Zr[idx_z] > 0:
                pairs.add((idx_z, idx_p))

    for idx_z, idx_p in pairs:
        if idx_z is None or idx_p is None:
            continue
        algorithms.resrank_program(
            aid.fitter,
            rank_zp_idx_list,
            "rr",
            ["ZrDelIdx", idx_z, "PrDelIdx", idx_p],
            marginalize_delay=marginalize_delay,
        )
    return rank_zp_idx_list


def ranking_reduction_c2r(
    aid,
    marginalize_delay=True,
):
    aid = ensure_aid(aid)
    rank_zp_idx_list = []

    Pc = aid.fitter.poles.c
    Zc = aid.fitter.zeros.c

    orders = aid.fitter_orders()
    rdmax = aid.hint("relative_degree_max")
    rdmin = aid.hint("relative_degree_min")

    if rdmin is None or orders.reldeg - 1 >= rdmin:
        for idx_z, z in enumerate(Zc):
            Q_rank = Q_rank_single(z)
            if Q_rank > 2:
                continue

            z_r = -abs(z)
            algorithms.resrank_program(
                aid.fitter,
                rank_zp_idx_list,
                "c2r",
                ["ZcDelIdx", idx_z, "ZrAdd", z_r],
                marginalize_delay=marginalize_delay,
            )

    if rdmax is None or orders.reldeg + 1 <= rdmax:
        for idx_p, p in enumerate(Pc):
            Q_rank = Q_rank_single(p)
            if Q_rank > 2:
                continue

            p_r = -abs(p)
            algorithms.resrank_program(
                aid.fitter,
                rank_zp_idx_list,
                "c2r",
                ["PcDelIdx", idx_p, "PrAdd", p_r],
                marginalize_delay=marginalize_delay,
            )

    return rank_zp_idx_list


def ranking_reduction_crr(
    aid,
    marginalize_delay=True,
):
    aid = ensure_aid(aid)
    rank_zp_idx_list = []

    Pc = aid.fitter.poles.c
    Zc = aid.fitter.zeros.c
    Pr = aid.fitter.poles.r
    Zr = aid.fitter.zeros.r

    Zc_argQsort = np.argsort(Q_rank_single(Zc))
    Pc_argQsort = np.argsort(Q_rank_single(Pc))

    Zr_BWsort = abs(Zr.real)
    Zr_argBWsort = np.argsort(Zr_BWsort)
    Zr_BWsort = Zr_BWsort[Zr_argBWsort]

    Pr_BWsort = abs(Pr.real)
    Pr_argBWsort = np.argsort(Pr_BWsort)
    Pr_BWsort = Pr_BWsort[Pr_argBWsort]

    def crr_ranker(
        Xc,
        Xc_argQsort,
        Yr,
        Yr_BWsort,
        Yr_argBWsort,
        Xc_code,
        Yr_code,
        include_unbalanced,
    ):
        if len(Yr_BWsort) <= 1:
            return
        for idx_x in Xc_argQsort[:3]:
            x = Xc[idx_x]
            bw_eff = abs(x)
            y_sort_idx = np.searchsorted(Yr_BWsort, bw_eff)

            if len(Yr_BWsort) >= 2:
                if y_sort_idx == 0:
                    idx_y1 = 0
                    idx_y2 = 1
                elif y_sort_idx >= len(Yr_BWsort) - 1:
                    idx_y1 = len(Yr_BWsort) - 1
                    idx_y2 = len(Yr_BWsort) - 2
                else:
                    idx_y1 = y_sort_idx
                    idx_y2 = y_sort_idx + 1
                # remap from sorted idx back into the original list idx
                idx_y1 = Yr_argBWsort[idx_y1]
                idx_y2 = Yr_argBWsort[idx_y2]
                # ensure that the ordering of the program is guaranteed
                if idx_y1 > idx_y2:
                    idx_y1, idx_y2 = idx_y2, idx_y1

                algorithms.resrank_program(
                    aid.fitter,
                    rank_zp_idx_list,
                    "crr",
                    [Xc_code, idx_x, Yr_code, idx_y1, Yr_code, idx_y2],
                    marginalize_delay=marginalize_delay,
                )

            if include_unbalanced and len(Yr_BWsort) >= 1:
                if y_sort_idx == 0:
                    idx_y = 0
                elif y_sort_idx >= len(Yr_BWsort) - 1:
                    idx_y = len(Yr_BWsort) - 1
                else:
                    if bw_eff - Yr[y_sort_idx] < Yr[y_sort_idx + 1] - bw_eff:
                        idx_y = y_sort_idx
                    else:
                        idx_y = y_sort_idx + 1
                # remap from sorted idx back into the original list idx
                idx_y = Yr_argBWsort[idx_y]

                algorithms.resrank_program(
                    aid.fitter,
                    rank_zp_idx_list,
                    "cr",
                    [Xc_code, idx_x, Yr_code, idx_y],
                    marginalize_delay=marginalize_delay,
                )

    orders = aid.fitter_orders()
    rdmax = aid.hint("relative_degree_max")
    rdmin = aid.hint("relative_degree_min")

    crr_ranker(
        Xc=Zc,
        Xc_argQsort=Zc_argQsort,
        Yr=Pr,
        Yr_BWsort=Pr_BWsort,
        Yr_argBWsort=Pr_argBWsort,
        Xc_code="ZcDelIdx",
        Yr_code="PrDelIdx",
        include_unbalanced=rdmin is None or orders.reldeg - 1 >= rdmin,
    )

    crr_ranker(
        Xc=Pc,
        Xc_argQsort=Pc_argQsort,
        Yr=Zr,
        Yr_BWsort=Zr_BWsort,
        Yr_argBWsort=Zr_argBWsort,
        Xc_code="PcDelIdx",
        Yr_code="ZrDelIdx",
        include_unbalanced=rdmax is None or orders.reldeg + 1 <= rdmax,
    )

    return rank_zp_idx_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np

from ..codings_cmn import (
    CodingTypeZ,
)


class CodingMOS(CodingTypeZ):
    # settable!
    N_parameters = None
    N_coeffs = 0
    p_coeffs = (1,)

    def setup(
        self,
        N_coeffs=None,
    ):
        if N_coeffs is not None:
            self.N_coeffs = N_coeffs
            coeffs = list(self.p_coeffs)[: self.N_parameters + 1]
            self.p_coeffs = np.array(coeffs + [0] * (1 + self.N_coeffs - len(coeffs)))

        N_parameters = self.N_coeffs
        self.N_parameters = N_parameters
        return

    def update(self, *coeffs):
        assert len(coeffs) == self.N_parameters
        self.p_coeffs = coeffs

    def reduce(self):
        return self.p_coeffs

    def update_roots(self, *rs):
        """
        r2, may be unspecified, in which case it is assumed to be nothing, if r1 is real, or otherwise the conjugate of r1
        """
        return

    def transfer(self):
        # multi-order sections (number indicates the order)
        Xn = self.sys.Xzn_grid
        Xn_power = Xn
        value = 1
        for cn in self.p_coeffs:
            value = value + cn * Xn_power
            Xn_power = Xn_power * Xn
        return value

    def derivative(self):
        # multi-order sections (number indicates the order)
        Xn = self.sys.Xzn_grid
        Xn_power = Xn
        jac_list = []
        for cn in self.p_coeffs:
            jac_list.append(Xn_power)
            Xn_power = Xn_power * Xn
        return jac_list

    def derivative_wtrans(self, sys):
        # multi-order sections (number indicates the order)
        Xn = sys.Xzn_grid
        Xn_power = Xn
        jac_list = []
        value = 1
        for cn in self.p_coeffs:
            value = value + cn * Xn_power
            jac_list.append(Xn_power)
            Xn_power = Xn_power * Xn
        return value, jac_list

    def roots(self):
        # multi-order sections (number indicates the order)
        return np.polynomial.polynomial.polyroots([1] + self.p_coeffs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np

from .base import (
    CodingType,
    Ipi,
    # I2pi
)

# import scipy.linalg


class CodingnlFA(CodingType):
    N_parameters = 2
    unstable = False
    p_nlF_Hz = 0
    p_amplitude = 0
    lock_nlF_Hz = False
    lock_amplitude = False
    hide_nlF_Hz = False
    hide_amplitude = False
    max_amplitude = 1

    def setup(
        self,
        hide_all=None,
        lock_nlF_Hz=None,
        hide_nlF_Hz=None,
        lock_amplitude=None,
        hide_amplitude=None,
        disable=None,
    ):
        if hide_all:
            hide_nlF_Hz = True
            hide_amplitude = True

        if hide_amplitude is not None:
            self.hide_amplitude = hide_amplitude

        N_parameters = 2
        if lock_nlF_Hz is not None:
            self.lock_nlF_Hz = lock_nlF_Hz
        if hide_nlF_Hz is not None:
            self.hide_nlF_Hz = hide_nlF_Hz
        if self.hide_nlF_Hz:
            N_parameters -= 1

        if lock_amplitude is not None:
            self.lock_amplitude = lock_amplitude
        if hide_amplitude is not None:
            self.hide_amplitude = hide_amplitude
        if self.hide_amplitude:
            N_parameters -= 1

        if disable is not None:
            self.disable = disable
        if self.disable:
            N_parameters = 0

        self.N_parameters = N_parameters

    def update(self, A=None, B=None):
        if self.disable:
            assert A is None and B is None
            return
        if self.hide_amplitude:
            assert B is None
            if self.hide_nlF_Hz:
                assert A is None
            else:
                nlF_Hz = A
                if not self.lock_nlF_Hz:
                    self.p_nlF_Hz = nlF_Hz
        else:
            if self.hide_nlF_Hz:

                amplitude = A
                if not self.lock_amplitude:
                    self.p_amplitude = amplitude
            else:
                nlF_Hz, amplitude = A, B
                if not self.lock_nlF_Hz:
                    self.p_nlF_Hz = nlF_Hz

                if not self.lock_amplitude:
                    self.p_amplitude = amplitude

    def reduce(self):
        if self.disable:
            return []
        if self.hide_amplitude:
            if self.hide_nlF_Hz:
                return []
            else:
                return [self.p_nlF_Hz]
        else:
            if self.hide_nlF_Hz:
                return [self.p_amplitude]
            else:
                return [
                    self.p_nlF_Hz,
                    self.p_amplitude,
                ]

    def option_set(self, minimum_BW_Hz=None, **kwargs):
        super(CodingnlFA, self).option_set(**kwargs)
        if minimum_BW_Hz is not None:
            self.max_amplitude = 1 - (minimum_BW_Hz / self.sys.F_nyquist_Hz)
        # TODO, should modify p_nl_A for the new max_amplitude
        return

    def update_roots(self, r1):
        """ """
        # TODO should honor max_amplitude
        amp = abs(r1)
        F_Hz = np.angle(r1) / np.pi * sys.F_nyquist_Hz
        relF_Hz = F_Hz / sys.F_cutoff_Hz
        if relF_Hz >= 1:
            relF_Hz = 0.999
        elif relF_Hz <= 0:
            relF_Hz = 0.001
        self.p_nlF_Hz = relF_Hz / (1 - relF_Hz ** 2) ** 0.5

        if amp < 1:
            # amp = x / (1. + x**2)**.5
            # amp**2 = x**2 / (1. + x**2)
            # amp**2 (1. + x**2) = x**2
            # amp**2  = x**2 ( 1 - amp**2)
            # amp**2 / (1 - amp**2) = x**2
            self.unstable = False
            self.p_amplitude = amp / (1 - amp ** 2) ** 0.5
        else:
            # amp = (1. + x**2)**.5 / x
            # amp**2 =  (1. + x**2) / x**2
            # amp**2 x**2 = x**2 + 1
            # amp**2  = x**2 (amp**2 - 1)
            # amp**2 / (amp**2 - 1) = x**2
            self.unstable = True
            self.p_amplitude = amp / (amp ** 2 - 1) ** 0.5
        return

    @property
    def relF_Hz(self):
        val = self.p_nlF_Hz / (1.0 + self.p_nlF_Hz ** 2) ** 0.5
        return val

    @property
    def amplitude(self):
        if not self.unstable:
            amp = (
                self.max_amplitude
                * self.p_amplitude
                / (1.0 + self.p_amplitude ** 2) ** 0.5
            )
        else:
            amp = (
                (1.0 + self.p_amplitude ** 2) ** 0.5 / self.p_amplitude
            ) / self.max_amplitude
        return amp

    def transfer(self):
        # frequency, logarithmic amplitude
        amp = self.amplitude
        r = amp * np.cos(np.pi * sys.F_cutoff_Hz * self.relF_Hz / sys.F_nyquist_Hz)
        Xn = sys.Xn_grid
        Xnsq = sys.Xn_grid_sq
        return (amp * amp) * Xnsq - 2 * Xn * r + 1

    def derivative(self):
        if self.disable:
            return []
        # real/imaginary part of root
        if not self.unstable:
            sqp5 = (1.0 + self.p_amplitude ** 2) ** 0.5
            amp = self.max_amplitude * self.p_amplitude / sqp5
            DampDl = self.max_amplitude * (
                1 / sqp5 - self.p_amplitude ** 2 / (1.0 + self.p_amplitude ** 2) ** 1.5
            )
        else:
            sqp5 = (1.0 + self.p_amplitude ** 2) ** 0.5
            amp = sqp5 / self.p_amplitude / self.max_amplitude
            DampDl = 1 / sqp5 - sqp5 / self.p_amplitude ** 2 / self.max_amplitude

        Fsqp5 = (1.0 + self.p_nlF_Hz ** 2) ** 0.5
        relF_Hz = self.p_nlF_Hz / Fsqp5
        DrelFHzDp = 1 / Fsqp5 - self.p_nlF_Hz ** 2 / (1.0 + self.p_nlF_Hz ** 2) ** 1.5

        F_nyquist_Hz = sys.F_nyquist_Hz
        Xn = sys.Xn_grid
        Xnsq = sys.Xn_grid_sq
        if not self.hide_nlF_Hz:
            if not self.hide_amplitude:
                rcos = np.cos(np.pi * sys.F_cutoff_Hz * relF_Hz / F_nyquist_Hz)
                rsin = np.sin(np.pi * sys.F_cutoff_Hz * relF_Hz / F_nyquist_Hz)
                return [
                    DrelFHzDp
                    * (2 * amp * np.pi * sys.F_cutoff_Hz / sys.F_nyquist_Hz)
                    * (Xn * rsin)
                    if not self.lock_nlF_Hz
                    else 0,
                    ((2 * amp * DampDl) * Xnsq - Xn * (2 * rcos * DampDl))
                    if not self.lock_amplitude
                    else 0,
                ]
            else:
                rsin = np.sin(np.pi * sys.F_cutoff_Hz * relF_Hz / F_nyquist_Hz)
                return [
                    DrelFHzDp
                    * (2 * amp * np.pi * sys.F_cutoff_Hz / sys.F_nyquist_Hz)
                    * (Xn * rsin)
                    if not self.lock_nlF_Hz
                    else 0,
                ]
        else:
            if not self.hide_amplitude:
                rcos = np.cos(np.pi * sys.F_cutoff_Hz * relF_Hz / F_nyquist_Hz)
                return [
                    ((2 * amp * DampDl) * Xnsq - Xn * (2 * rcos * DampDl))
                    if not self.lock_amplitude
                    else 0,
                ]
            else:
                return []

    def roots_c(self):
        # real/imaginary part of root
        return [
            self.amplitude
            * np.exp(Ipi * abs(self.relF_Hz) * sys.F_cutoff_Hz / sys.F_nyquist_Hz)
        ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
from wield.bunch import Bunch
from . import codings_s
from . import codings_z

_common = dict()

coding_maps = Bunch(
    RI=Bunch(
        rep=Bunch(
            Sf=codings_s.coding_maps.RI,
            Z=codings_z.coding_maps.RI,
        ),
        **_common
    ),
    FBW=Bunch(
        rep=Bunch(
            Sf=codings_s.coding_maps.FBW,
            Z=codings_z.coding_maps.FBW,
        ),
        **_common
    ),
    nlFBW=Bunch(
        rep=Bunch(
            Sf=codings_s.coding_maps.nlFBW,
            Z=codings_z.coding_maps.nlFBW,
        ),
        **_common
    ),
    nlFBW_safe=Bunch(
        rep=Bunch(
            Sf=codings_s.coding_maps.nlFBW_safe,
            Z=codings_z.coding_maps.nlFBW_safe,
        ),
        **_common
    ),
    SOS=Bunch(
        rep=Bunch(
            Sf=codings_s.coding_maps.SOS,
            Z=codings_z.coding_maps.SOS,
        ),
        **_common
    ),
    # SOS_safe = Bunch(
    #    rep = Bunch(
    #        Sf = codings_s.coding_maps.SOS_safe,
    #        Z  = codings_z.coding_maps.SOS_safe,
    #    ),
    #    **_common
    # ),
)



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

from wield.iirrational import v2
from wield.iirrational.v2 import testing
from wield.iirrational import plots


try:
    from IIRrational_test_data import IIRrational_data_dev
except ImportError:
    module_import_skip = True
else:
    module_import_skip = False


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_quad_M0_P2V_10Hz_S(request, browser, plotsections, plot_verbosity):
    data_name = "quad_M0_P2V_10Hz"

    dataB = IIRrational_data_dev(
        data_name,
        instance_num=1,
        set_num=2,
    )

    out = v2.data2filter(
        F_Hz=dataB.F_Hz,
        data=dataB.data,
        SNR=dataB.SNR,
        F_nyquist_Hz=None,
        # order_initial = 80,
        total_degree_min=None,
        delay_s=0,
        mode="rational",
        baseline_only=True,
        # SNR_cuttoff = 50,
        hints=[
            testing.validate_plot_log(__file__, request),
        ],
    )

    # out.fitter.matched_pairs_clear(Q_rank_cutoff = .4)
    def plot():
        return plots.plot_fit(
            out.fitter,
            plot_zp=True,
            # plot_past_data = False,
            plot_past_data=True,
            xscale="log",
        )

    with testing.plot_on_assert(__file__, request, plot, plot_anyway=True):
        pass
        # assert(False)
    # with plot_on_assert(__file__, request, out.fitter, plot_anyway = True):
    #    pass
    return


# @pytest.mark.skipif(module_import_skip, reason = 'cannot import IIRrational_test_data')
# def test_quad_M0_P2V_10Hz_rat_Z(browser, plotsections, plot_verbosity):
#    data_name = 'quad_M0_P2V_10Hz'
#    out = v1.rational_disc_fit(
#        IIRrational_data_dev(
#            data_name,
#            instance_num = 1,
#            set_num = 2,
#        ),
#        nyquist_final_Hz = 1e3,
#        order = 100,
#    )
#
#    import matplotlib.pyplot as plt
#    from wield.iirrational import plots
#    axB = plots.plot_fitter_flag(
#        out,
#        #don't use jpg as matplotlib can't write it out of the box (needs pillow)
#        fname = path.join(path.split(__file__)[0], 'out-cplx/test-rational-Z.png'),
#    )
#    plt.close(axB.fig)
#    plt.show()
#    return
#
# @pytest.mark.skipif(module_import_skip, reason = 'cannot import IIRrational_test_data')
# def test_quad_M0_P2V_10Hz_rat_S(browser, plotsections, plot_verbosity):
#    data_name = 'quad_M0_P2V_10Hz'
#    out = v1.rational_disc_fit(
#        IIRrational_data_dev(
#            data_name,
#            instance_num = 1,
#            set_num = 2,
#        ),
#        nyquist_final_Hz = 0,
#        order = 100,
#    )
#
#    import matplotlib.pyplot as plt
#    from wield.iirrational import plots
#    axB = plots.plot_fitter_flag(
#        out,
#        #don't use jpg as matplotlib can't write it out of the box (needs pillow)
#        fname = path.join(path.split(__file__)[0], 'out-cplx/test-rational-S.png'),
#    )
#    plt.close(axB.fig)
#    plt.show()
#    return



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest
import numpy as np
import os.path as path

from wield.iirrational.testing import IIRrational_data
from wield.iirrational import file_io
from wield.iirrational.file_io import (
    matlab_io,
)

from wield.iirrational.utilities.print import pprint

try:
    from IIRrational_test_data import matlab as matlab_test_data
except ImportError:
    module_import_skip = True
else:
    module_import_skip = False


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_mat():
    fname = matlab_test_data.matfiles["iir_test_vars.mat"]
    fdict = matlab_io.load_matlab(fname)
    fdict["F_Hz"]
    assert fdict["xfer"].dtype == np.complex
    fdict["SNR"]
    assert fdict["noneA"] is None
    assert fdict["noneC"] is None


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_mat_struct():
    fname = matlab_test_data.matfiles["iir_test_struct.mat"]
    fdict = matlab_io.load_matlab(fname)
    # pprint(fdict['data']['xfer'])
    assert file_io.subkey_search(fdict, "data.xfer").dtype == complex
    # pprint(fdict['data']['confuse'])
    assert fdict["data"]["noneA"] is None
    assert fdict["data"]["noneC"] is None


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_mat_struct_RI():
    fname = matlab_test_data.matfiles["iir_test_struct_RI.mat"]
    fdict = matlab_io.load_matlab(fname)
    # pprint(fdict)


@pytest.mark.skipif(module_import_skip, reason="cannot import IIRrational_test_data")
def test_mat_struct_RT():
    fname = matlab_test_data.matfiles["iir_test_struct.mat"]
    fdict = matlab_io.load_matlab(fname)
    test_path = path.split(__file__)[0]
    fpath = path.join(test_path, "iir_test_struct_RT.mat")
    matlab_io.save_matlab(fpath, fdict)
    fdict2 = matlab_io.load_matlab(fpath)

    assert np.all(fdict["data"]["F_Hz"] == fdict2["data"]["F_Hz"])
    assert np.all(fdict["data"]["xfer"] == fdict2["data"]["xfer"])
    assert np.all(fdict["data"]["SNR"] == fdict2["data"]["SNR"])
    assert np.all(fdict["data"]["noneA"] == fdict2["data"]["noneA"])
    assert np.all(fdict["data"]["noneC"] == fdict2["data"]["noneC"])
    print(str(fdict["data"]["confuse"]))
    print(str(fdict2["data"]["confuse"]))
    assert str(fdict["data"]["confuse"]) == str(fdict2["data"]["confuse"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import contextlib
import numpy as np

from .arguments import ArgumentError
from .. import fitters_ZPK
import warnings
import collections

from ..TFmath import match_SOS_pairs
from .. import representations


class ResultsAid(object):
    def __init__(self, fit_aid, kw=None):
        self._current_order = None
        self._fitter = None
        self.kw = kw

        self.initial_arguments = {}
        self.fitting_function = None

        self.fit_aid = fit_aid

        self.__fitters_by_order = None
        self._extra_data_by_order = dict()
        return

    @property
    def fitter(self):
        return self._fitter

    @property
    def _fitter_extra(self):
        extra = self._extra_data_by_order.setdefault(self._current_order, {})
        return extra

    @property
    def fitterRI(self):
        extra = self._fitter_extra
        fitterRI = extra.get("fitterRI", None)
        if fitterRI is None:
            fitterRI = self.fitter.regenerate(
                ZPKrep=self.fitter.ZPKrep,
                coding_map=fitters_ZPK.coding_maps.RI
            )
            extra["fitterRI"] = fitterRI
        return fitterRI

    @property
    def _fitters_by_order(self):
        """
        Returns four arrays, an array of fitter orders, residuals, the
        fitters themselves, and a bool array indicating if each fitter is the
        local best (the fitter of lowest residuals with equal or lower order).
        They are sorted by the fitter order such that binary
        search may be performed on the order array.
        """
        if self.__fitters_by_order is None:
            fitters_by_order = {}
            for fmeta in self.fit_aid._fitters:
                if not fmeta.valid:
                    continue

                fgroup = fitters_by_order.setdefault(fmeta.fitter.order, [])
                fgroup.append(fmeta.fitter)
            sorted_orders = []
            sorted_fitters = []
            sorted_residuals = []
            sorted_local_best = []
            lowest_res = float("inf")
            for order, fgroup in sorted(fitters_by_order.items()):
                fgroup.sort(key=lambda f: f.residuals_average)
                sorted_orders.append(order)
                sorted_fitters.append(fgroup[0])
                resavg = fgroup[0].residuals_average
                sorted_residuals.append(resavg)
                if resavg < lowest_res:
                    lowest_res = resavg
                    sorted_local_best.append(True)
                else:
                    sorted_local_best.append(False)
            sorted_local_best = np.asarray(sorted_local_best)
            self.__fitters_by_order = (
                np.asarray(sorted_orders),
                np.asarray(sorted_residuals),
                # currently a bug in declarative that attribute misses cause RuntimeError rather than AttributeError
                # numpy now duck-type checks this, and the wrong exception is failing. Instead, casting to list
                list(sorted_fitters),
                np.argwhere(sorted_local_best)[:, 0],
            )
        return self.__fitters_by_order

    @property
    def order(self):
        return self._current_order

    def residuals_by_order(self):
        obo, rbo, fbo, lbo = self._fitters_by_order
        return obo, rbo

    def choose(self, order=None, direction=None, **kwargs):
        """
        Chooses the filter with the lowest residual of order equal or less than
        requested. The data2filter algorithm will already have chosen a filter,
        but this allows investigating alternatives among the ones explored by
        the algorithm.

        if order is None, then the order is taken to be the current order.

        Can take argument "direction" to specify which nearest order to find
         - "<=" or "le": less than or equal to the specified order
         - ">=" or "ge": greater than or equal to the specified order
         - "<" or "l": less than the specified order
         - ">" or "g": greater than or equal to the specified order

        Note that calling
        results.choose(direction = '>') will return the fitter with the next
        order above the current one, and calling results.choose(direction = '<')
        gives the next lower order. So that one can investigate the nearest
        alternative choices.
        """
        if direction is None:
            direction = "<="

        if order is None:
            order = self.order

        dmap = {
            "ge": ">=",
            "g": ">",
            ">=": ">=",
            ">": ">",
            "le": "<=",
            "l": "<",
            "<=": "<=",
            "<": "<",
        }

        try:
            direction = dmap[direction]
        except KeyError:
            raise ArgumentError(
                ("'direction' argument must be one of {}").format(dmap.keys)
            )

        # using the while just as a goto statement with 'break' keyword
        while True:
            obo, rbo, fbo, lbo = self._fitters_by_order
            idx_ord = np.searchsorted(obo, order)
            if idx_ord == 0:
                if direction in ("<=", "<"):
                    break
            elif idx_ord >= len(obo):
                if direction in (">=", ">"):
                    idx_ord = len(obo) - 1
                    break
                idx_ord = len(obo) - 1

            if order == obo[idx_ord]:
                if direction == "<":
                    idx_ord -= 1
                elif direction == ">":
                    idx_ord += 1

            if direction in ("<", "<="):
                idx_ord = np.argmin(rbo[: idx_ord + 1])
            break
        fitter = fbo[idx_ord]

        self._fitter = fitter
        self._current_order = obo[idx_ord]
        return fitter

    @contextlib.contextmanager
    def choice(self, order=None, direction=None, **kwargs):
        """
        Allows a temporary choice of fitter, for investigations without having
        to store the optimal choice. Takes the same arguments as choose().
        """
        fitter_prev = self._fitter
        self.choose(order=order, direction=direction, **kwargs)
        yield
        self._fitter = fitter_prev
        return

    def as_scipy_signal_ZPKsw(self, with_delay=False):
        """
        Returns the chosen filter in scipy format
        """
        # TODO
        from ..LIGO.conversion import filter2sortedZPK

        z, p, k = filter2sortedZPK(self.fitter)
        z = np.asarray(z) * 2 * np.pi
        p = np.asarray(p) * 2 * np.pi
        k = k * (2 * np.pi) ** (len(p) - len(z))

        if not with_delay:
            if self.fitter.delay_s != 0:
                warnings.warn(
                    (
                        "fit export not including delay,"
                        " but filter has delay {}s. Call results.as_scipy_signal_ZPKsw(with_delay = True)"
                        " and the output will be ZPKD"
                    ).format(self.fitter.delay_s)
                )
            return z, p, k
        else:
            return z, p, k, self.fitter.delay_s

    def as_fitterMRF_ZPKz(
        self,
        t_sample_s=None,
        F_sample_Hz=None,
        F_nyquist_Hz=None,
    ):
        """
        Returns the chosen filter in scipy format for the z domain
        (must specify t_sample_s, or F_sample_Hz or F_nyquist_Hz). This
        transforms into the Z domain, then reruns the optimization to ensure a
        best fit in the transformed coordinates.
        """
        F_nyquist_Hz = argF_nyquist_Hz(
            t_sample_s=t_sample_s,
            F_sample_Hz=F_sample_Hz,
            F_nyquist_Hz=F_nyquist_Hz,
        )
        fitter_z = self._fitter_extra.get(F_nyquist_Hz, None)
        if fitter_z is None:
            xrep = self.fitter.ZPKrep.change_F_Nyquist_Hz(F_nyquist_Hz)
            raise NotImplementedError()
            self._fitter_extra[F_nyquist_Hz] = fitter_z
        return fitter_z

    def as_scipy_signal_ZPKz(
        self,
        t_sample_s=None,
        F_sample_Hz=None,
        F_nyquist_Hz=None,
    ):
        """
        Returns the chosen filter in scipy format for the z domain
        (must specify t_sample_s, or F_sample_Hz or F_nyquist_Hz). This
        transforms into the Z domain, then reruns the optimization to ensure a
        best fit in the transformed coordinates.
        """
        fitter_z = self.as_fitterMRF_ZPKz(
            t_sample_s=t_sample_s,
            F_sample_Hz=F_sample_Hz,
            F_nyquist_Hz=F_nyquist_Hz,
        )
        raise NotImplementedError()
        return

    def as_ZPKrep(
        self,
    ):
        """
        Returns the filter as an wield.iirrational ZPKwData object, for use with other
        fitters in the wield.iirrational package.
        """
        return self.fitter.ZPKrep

    @property
    def ZPKrep(self):
        return self.as_ZPKrep()

    def as_ZPKTF(self, delay2ZPK_args):
        """
        Returns the filter as an wield.iirrational ZPKTF object, for use
        with mathematical analysis. Note, if the delay is nonzero,
        some conversion of the delay into ZPK is required. If the delay is
        positive (as is usual), the some order of RHP-zero/LHP-pole pairs must
        be added to approximate the delay.
        """
        # TODO
        raise NotImplementedError()

    def as_matlab_str_ZPKsw(
        self,
    ):
        """
        Returns the chosen filter in matlab format for the z domain
        (must specify t_sample_s, or F_sample_Hz or F_nyquist_Hz)
        """
        # TODO
        raise NotImplementedError()

    def as_foton_ZPKsf(
        self,
        with_delay=False,
    ):
        """
        Returns the chosen filter as a string foton formatting. Uses the
        frequency domain representation
        """
        from ..LIGO.conversion import filter2fotonZPK

        ZPK = filter2fotonZPK(
            self.fitter.ZPKrep,
            plane="f",
            zpk_output=True,
        )
        if not with_delay:
            if self.fitter.delay_s != 0:
                warnings.warn(
                    (
                        "fit export not including delay,"
                        " but filter has delay {}s. Call results.as_foton_ZPKsf(with_delay = True)"
                        " and the output will be ZPKD"
                    ).format(self.fitter.delay_s)
                )
            return ZPK
        else:
            return tuple(ZPK) + (self.fitter.delay_s,)

    def as_foton_ZPKsn(
        self,
        with_delay=False,
    ):
        """
        Returns the chosen filter as a string foton formatting. Uses the
        frequency domain representation
        """
        from ..LIGO.conversion import filter2fotonZPK

        ZPK = filter2fotonZPK(
            self.fitter.ZPKrep,
            plane="n",
            zpk_output=True,
        )
        if not with_delay:
            if self.fitter.delay_s != 0:
                warnings.warn(
                    (
                        "fit export not including delay,"
                        " but filter has delay {}s. Call results.as_foton_ZPKsn(with_delay = True)"
                        " and the output will be ZPKD"
                    ).format(self.fitter.delay_s)
                )
            return ZPK
        else:
            return tuple(ZPK) + (self.fitter.delay_s,)

    def as_refine_str(self, sigfigs=5):
        zzpp_pairs = match_SOS_pairs(
            self.fitter.zeros.r,
            self.fitter.zeros.c,
            self.fitter.poles.r,
            self.fitter.poles.c,
            F_nyquist_Hz=None,
        )
        zs = []
        ps = []

        def append(rs, r):
            if r is not None:
                if r.imag != 0:
                    if r.imag > 0:
                        rs.append(
                            "{real:.{sf}f}+{imag:.{sf}f}j".format(
                                real=r.real, imag=r.imag, sf=sigfigs
                            )
                        )
                    else:
                        rs.append(
                            "{real:.{sf}f}-{imag:.{sf}f}j".format(
                                real=r.real, imag=-r.imag, sf=sigfigs
                            )
                        )
                else:
                    rs.append("{real:.{sf}f}".format(real=r.real, sf=sigfigs))

        for (z1, z2, p1, p2) in zzpp_pairs:
            append(zs, z1)
            append(zs, z2)
            append(ps, p1)
            append(ps, p2)
        return "--zeros={} --poles={}".format(",".join(zs), ",".join(ps))

    def as_refine_str(self, sigfigs=5):
        zs_r = []
        ps_r = []
        zs_c = []
        ps_c = []

        def append(rs, r):
            if r is not None:
                if r.imag != 0:
                    if r.imag > 0:
                        rs.append(
                            "{real:.{sf}f}+{imag:.{sf}f}j".format(
                                real=r.real, imag=r.imag, sf=sigfigs
                            )
                        )
                    else:
                        rs.append(
                            "{real:.{sf}f}-{imag:.{sf}f}j".format(
                                real=r.real, imag=-r.imag, sf=sigfigs
                            )
                        )
                else:
                    rs.append("{real:.{sf}f}".format(real=r.real, sf=sigfigs))

        for r in sorted(self.fitter.zeros.r, key=lambda r: -r.real):
            append(zs_r, r)
        for r in sorted(self.fitter.poles.r, key=lambda r: -r.real):
            append(ps_r, r)

        for r in sorted(self.fitter.zeros.c, key=lambda r: (-np.sign(r.real), r.imag)):
            append(zs_c, r)
        for r in sorted(self.fitter.poles.c, key=lambda r: (-np.sign(r.real), r.imag)):
            append(ps_c, r)

        return "--zeros={}:{} --poles={}:{}".format(
            ",".join(zs_r),
            ",".join(zs_c),
            ",".join(ps_r),
            ",".join(ps_c)
        )

    def as_foton_str_ZPKsf(
        self,
        annotate_pairs=False,
    ):
        """
        Returns the chosen filter as a string foton formatting. Uses the
        frequency domain representation
        """
        from ..LIGO.conversion import filter2fotonZPK

        return filter2fotonZPK(
            self.fitter.ZPKrep,
            annotate_pairs=annotate_pairs,
            plane="f",
        )

    def as_foton_str_ZPKsn(
        self,
        annotate_pairs=False,
    ):
        """
        Returns the chosen filter as a string foton formating. Uses the
        "normalized" frequency domain representation.
        """
        # TODO
        from ..LIGO.conversion import filter2fotonZPK

        return filter2fotonZPK(
            self.fitter.ZPKrep,
            annotate_pairs=annotate_pairs,
            plane="n",
        )

    def as_foton_str_ZPKz(
        self,
        t_sample_s=None,
        F_sample_Hz=None,
        F_nyquist_Hz=None,
    ):
        """
        Returns the chosen filter in foton format native for the z domain
        (must specify t_sample_s, or F_sample_Hz or F_nyquist_Hz). This
        transforms into the Z domain, then reruns the optimization to ensure a
        best fit in the transformed coordinates.
        """
        # TODO
        from ..LIGO.conversion import filter2fotonZPK

        fitter_z = self.as_fitterMRF_ZPKz(
            t_sample_s=t_sample_s,
            F_sample_Hz=F_sample_Hz,
            F_nyquist_Hz=F_nyquist_Hz,
        )
        return filter2fotonZPK(fitter_z.ZPKrep, plane="z")

    def _fitter_export_dict(self, fitter):
        from ..LIGO.conversion import filter2fotonZPK

        zpk = tuple(fitter.ZPKrep.ZPK)
        d = collections.OrderedDict(
            [
                (
                    "zpk",
                    collections.OrderedDict(
                        [
                            ("z", zpk[0]),
                            ("p", zpk[1]),
                            ("k", zpk[2]),
                        ]
                    ),
                ),
                ("delay_s", fitter.ZPKrep.delay_s),
                ("gain", fitter.gain),
                (
                    "poles",
                    collections.OrderedDict(
                        [
                            ("r", fitter.poles.r),
                            ("c", fitter.poles.c),
                        ]
                    ),
                ),
                (
                    "zeros",
                    collections.OrderedDict(
                        [
                            ("r", fitter.zeros.r),
                            ("c", fitter.zeros.c),
                        ]
                    ),
                ),
                (
                    "poles_overlay",
                    collections.OrderedDict(
                        [
                            ("r", fitter.poles_overlay.r),
                            ("c", fitter.poles_overlay.c),
                        ]
                    ),
                ),
                (
                    "zeros_overlay",
                    collections.OrderedDict(
                        [
                            ("r", fitter.zeros_overlay.r),
                            ("c", fitter.zeros_overlay.c),
                        ]
                    ),
                ),
                ("residuals", fitter.residuals_average),
                ("residuals_max", fitter.residuals_max),
                ("residuals_med", fitter.residuals_med),
                (
                    "foton_Sf",
                    filter2fotonZPK(
                        fitter.ZPKrep,
                        annotate_pairs=False,
                        plane="f",
                    ),
                ),
            ]
        )
        return d

    def choice_export_dict(self):
        return self._fitter_export_dict(self.fitter)

    def alternatives_export_dict(self, local_best=True):
        obo, rbo, fbo, lbo = self._fitters_by_order

        # only returns or displays those which are the best for that order or below
        if local_best:
            obo = obo[lbo]
            rbo = rbo[lbo]
            fbo = [fbo[idx] for idx, val in enumerate(lbo) if val]

        exd = collections.OrderedDict()
        for order, fit in zip(obo, fbo):
            key = "order_{:0>2d}".format(order)
            exd[key] = self._fitter_export_dict(fit)

        return exd


def argF_nyquist_Hz(
    t_sample_s=None,
    F_sample_Hz=None,
    F_nyquist_Hz=None,
):
    if t_sample_s is not None:
        _F_nyquist_Hz = 1 / (2 * t_sample_s)
        if F_nyquist_Hz is not None and _F_nyquist_Hz != F_nyquist_Hz:
            raise ArgumentError(
                "t_sample_s and F_nyquist_Hz both" " given, but not consistent"
            )
        F_nyquist_Hz = _F_nyquist_Hz
    if F_sample_Hz is not None:
        _F_nyquist_Hz = F_sample_Hz / 2
        if F_nyquist_Hz is not None and _F_nyquist_Hz != F_nyquist_Hz:
            raise ArgumentError(
                "F_sample_Hz and F_nyquist_Hz both" " given, but not consistent"
            )
        F_nyquist_Hz = _F_nyquist_Hz
    return F_nyquist_Hz



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
import os
import re

from wield.bunch import Bunch


ext2type = {
    ".h5": "hdf5",
    ".hdf": "hdf5",
    ".hdf5": "hdf5",
    ".pkl": "pickle",
    ".pickle": "pickle",
    ".yaml": "yaml",
    ".yml": "yaml",
    ".json": "json",
    ".ini": "ini",
    ".mat": "mat",
    ".m": "mat",
    ".txt": "csv",
    ".csv": "csv",
    ".txt.gz": "csv",
    ".csv.bz2": "csv",
}


type2type = {
    "h5": "hdf5",
    "hdf": "hdf5",
    "hdf5": "hdf5",
    "yaml": "yaml",
    "yml": "yaml",
    "json": "json",
    "ini": "ini",
    "matlab": "mat",
    "mat": "mat",
    "m": "mat",
    "csv": "csv",
    "pkl": "pickle",
    "pickle": "pickle",
    "special": "special",
}


type2features = {
    "hdf5": dict(
        data=True,
        config=True,
        compact=True,
        complex=True,
        ndarray=True,
    ),
    "yaml": dict(
        data=True,
        config=True,
        compact=False,
        complex=False,
        ndarray=False,
    ),
    "json": dict(
        data=True,
        config=False,
        compact=False,
        complex=False,
        ndarray=False,
    ),
    "ini": dict(
        data=False,
        config=True,
        compact=False,
        complex=False,
        ndarray=False,
    ),
    "mat": dict(
        data=True,
        config=True,
        compact=True,
        complex=True,
        ndarray=True,
    ),
    "csv": dict(
        data=True,
        config=False,
        compact=True,
        complex=False,
        ndarray=True,
    ),
    "pickle": dict(
        data=True,
        config=True,
        compact=True,
        complex=True,
        ndarray=True,
    ),
}

re_FILEKEY = re.compile(r"(.*)\[(.*)\]")


def determine_type(fname):
    fspl = fname.split("::")

    if len(fspl) == 1:
        (fname,) = fspl
        m = re_FILEKEY.match(fname)
        if m:
            fname = m.group(1)
            subkey = m.group(2)
        else:
            subkey = None
        fbase, fext = os.path.splitext(fname)
        if fname[0] == ":" and fname[-1] == ":":
            ftype = "special"
            fname = fname[1:-1]
        else:
            ftype = ext2type[fext.lower()]
    elif len(fspl) == 2:
        fname, ftype = fspl
        if m:
            fname = m.group(1)
            subkey = m.group(2)
        else:
            subkey = None
        ftype = type2type[ftype.lower()]
        if ftype == "special" and fname[0] == ":" and fname[-1] == ":":
            fname = fname[1:-1]
    else:
        # TODO, be more specific about which argument caused this
        raise RuntimeError("Too many '::' in filetype specification")

    if ftype != "special":
        fname = os.path.abspath(fname)

    return Bunch(
        fname=fname,
        subkey=subkey,
        ftype=ftype,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np
from wield.bunch import Bunch


def BWz(r, F_nyquist_Hz):
    return (1 - r) * F_nyquist_Hz


def BWz_neg(r, F_nyquist_Hz):
    return (r - 1) * F_nyquist_Hz


def get_aspect(ax):
    figW, figH = ax.get_figure().get_size_inches()
    # Axis size on figure
    _, _, w, h = ax.get_position().bounds
    return (figW * w) / (figH * h)


def ZPKrep2Sf(ZPKrep):
    if ZPKrep.F_nyquist_Hz is None:
        return ZPKrep
    else:
        # TODO, make this use some other standard function
        def remap(rB):
            rB = rB.copy()
            select_r = rB.r > 0
            rB.r = (rB.r[select_r] - 1) * ZPKrep.F_nyquist_Hz
            F_Hz = np.angle(rB.c) / np.pi * ZPKrep.F_nyquist_Hz
            BW = (abs(rB.c) - 1) * ZPKrep.F_nyquist_Hz
            rB.c = BW + 1j * F_Hz
            rB.clear()
            return rB

    ret = Bunch(
        zeros=remap(ZPKrep.zeros),
        poles=remap(ZPKrep.poles),
        zeros_overlay=remap(ZPKrep.zeros_overlay),
        poles_overlay=remap(ZPKrep.poles_overlay),
        F_nyquist_Hz=None,
    )
    return ret
